"""add v2 role mappings

Revision ID: 771180018e1b
Revises: 30b37e725c32
Create Date: 2023-11-29 09:02:35.835664

"""

import sqlalchemy as sa

from alembic import op


# revision identifiers, used by Alembic.
revision = "771180018e1b"
down_revision = "30b37e725c32"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "namespace_role_mapping_v2",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("namespace_id", sa.Integer(), nullable=False),
        sa.Column("other_namespace_id", sa.Integer(), nullable=False),
        sa.Column("role", sa.Unicode(length=255), nullable=False),
        sa.ForeignKeyConstraint(
            ["namespace_id"],
            ["namespace.id"],
        ),
        sa.ForeignKeyConstraint(
            ["other_namespace_id"],
            ["namespace.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("namespace_id", "other_namespace_id", name="_uc"),
    )


def downgrade():
    op.drop_table("namespace_role_mapping_v2")



"""role mapping

Revision ID: b387747ca9b7
Revises: abd7248d5327
Create Date: 2023-07-04 14:35:48.177574

"""

import sqlalchemy as sa

from alembic import op


# revision identifiers, used by Alembic.
revision = "b387747ca9b7"
down_revision = "abd7248d5327"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "namespace_role_mapping",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("namespace_id", sa.Integer(), nullable=False),
        sa.Column("entity", sa.Unicode(length=255), nullable=False),
        sa.Column("role", sa.Unicode(length=255), nullable=False),
        sa.ForeignKeyConstraint(
            ["namespace_id"],
            ["namespace.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )

    op.add_column("namespace", sa.Column("metadata_", sa.JSON(), nullable=True))


def downgrade():
    op.drop_column("namespace", "metadata_")
    op.drop_table("namespace_role_mapping")



import os
import pathlib
import sys
import tempfile
import warnings

from typing import Union

import yaml

from conda_store_server import action, schema
from conda_store_server.action.utils import logged_command


def get_installer_platform():
    # This is how the default platform name is generated internally by
    # constructor. For example: osx-arm64, linux-64, win-64.
    # https://github.com/conda/constructor/blob/main/CONSTRUCT.md#available-platforms
    # Note: constructor is cross-friendly, see:
    # https://github.com/conda-incubator/conda-store/pull/714#discussion_r1465115323
    from conda.base.context import context

    return context.subdir


@action.action
def action_generate_constructor_installer(
    context,
    conda_command: str,
    specification: Union[schema.CondaSpecification, schema.LockfileSpecification],
    installer_dir: pathlib.Path,
    version: str,
    is_lockfile: bool = False,
):
    def write_file(filename, s):
        with open(filename, "w") as f:
            context.log.info(f"{filename}:\n{s}")
            f.write(s)

    # Checks if constructor is available
    try:
        command = [
            "constructor",
            "--help",
        ]
        logged_command(context, command)
    except FileNotFoundError:
        warnings.warn(
            "Installer generation requires constructor: https://github.com/conda/constructor"
        )
        return

    # pip dependencies are not directly supported by constructor, they will be
    # installed via the post_install script:
    # https://github.com/conda/constructor/issues/515
    # conda and pip need to be in dependencies for the post_install script
    dependencies = ["conda", "pip"]
    pip_dependencies = []

    if is_lockfile:
        # Adds channels
        channels = [c.url for c in specification.lockfile.metadata.channels]

        # Adds dependencies
        for p in specification.lockfile.package:
            # Ignores packages not matching the current platform. Versions can
            # be different between platforms or a package might not support all
            # platforms. constructor is cross-friendly, but we're currently
            # building only for the current architecture, see the comment in
            # get_installer_platform
            if p.platform not in ["noarch", get_installer_platform()]:
                continue
            if p.manager == "pip":
                pip_dependencies.append(f"{p.name}=={p.version}")
            else:
                ext = ".tar.bz2" if p.url.endswith(".tar.bz2") else ".conda"
                build_string = p.url[: -len(ext)].rsplit("-", maxsplit=1)[-1]
                dependencies.append(f"{p.name}=={p.version}={build_string}")

    else:
        # Adds channels
        channels = specification.channels

        # Adds dependencies
        for d in specification.dependencies:
            if isinstance(d, schema.CondaSpecificationPip):
                pip_dependencies.extend(d.pip)
            else:
                dependencies.append(d)

    # Creates the construct.yaml file and post_install script
    ext = ".exe" if sys.platform == "win32" else ".sh"
    pi_ext = ".bat" if sys.platform == "win32" else ".sh"
    installer_filename = (installer_dir / specification.name).with_suffix(ext)

    os.makedirs(installer_dir, exist_ok=True)

    with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as tmp_dir:
        tmp_dir = pathlib.Path(tmp_dir)
        cache_dir = tmp_dir / "pkgs"
        tmp_dir /= "build"
        os.makedirs(cache_dir, exist_ok=True)
        os.makedirs(tmp_dir, exist_ok=True)
        construct_file = tmp_dir / "construct.yaml"
        post_install_file = (tmp_dir / "post_install").with_suffix(pi_ext)

        construct = {
            "installer_filename": str(installer_filename),
            "post_install": str(post_install_file),
            "name": specification.name,
            "channels": channels,
            "specs": dependencies,
            "version": version,
        }

        if sys.platform == "win32":
            post_install = """\
call "%PREFIX%\Scripts\\activate.bat"
"""
        else:
            post_install = """\
#!/usr/bin/env bash
set -euxo pipefail
source "$PREFIX/etc/profile.d/conda.sh"
conda activate "$PREFIX"
"""
        if pip_dependencies:
            post_install += f"""
python -m pip install {' '.join(pip_dependencies)}
"""

        # Writes files to disk
        write_file(construct_file, yaml.dump(construct))
        write_file(post_install_file, post_install)

        # Calls constructor
        # Note: `cache_dir` is the same as the conda `pkgs` directory. It needs
        # to be specified here because the default `pkgs` directory is not
        # available in Docker, which was causing conda's `create_cache_dir` to
        # fail.
        command = [
            "constructor",
            "-v",
            "--cache-dir",
            str(cache_dir),
            "--platform",
            get_installer_platform(),
            str(tmp_dir),
        ]
        logged_command(context, command)

    return installer_filename



import pathlib
import shutil

from conda_store_server import action, conda_utils


@action.action
def action_remove_conda_prefix(
    context,
    conda_prefix: pathlib.Path,
):
    # safeguard to try and prevent destructive actions
    if not conda_utils.is_conda_prefix(conda_prefix):
        raise ValueError("given prefix is not a conda environment")

    context.log.info(f'removing conda prefix "{conda_prefix.resolve()}"')
    shutil.rmtree(conda_prefix)



import json
import typing
import uuid

from sqlalchemy.orm import Session

from conda_store_server import api, conda_utils, orm, schema


def seed_conda_store(
    db: Session,
    conda_store,
    config: typing.Dict[str, typing.Dict[str, schema.CondaSpecification]] = {},
):
    for namespace_name in config:
        namespace = api.ensure_namespace(db, name=namespace_name)
        for environment_name, specification in config[namespace_name].items():
            environment = api.ensure_environment(
                db,
                name=specification.name,
                namespace_id=namespace.id,
            )
            specification = api.ensure_specification(db, specification)
            build = api.create_build(db, environment.id, specification.id)
            db.commit()

            environment.current_build_id = build.id
            db.commit()

            _create_build_artifacts(db, conda_store, build)
            _create_build_packages(db, conda_store, build)

            api.create_solve(db, specification.id)
            db.commit()


def _create_build_packages(db: Session, conda_store, build: orm.Build):
    channel_name = conda_utils.normalize_channel_name(
        conda_store.conda_channel_alias, "conda-forge"
    )
    channel = api.ensure_conda_channel(db, channel_name)

    conda_package = orm.CondaPackage(
        name=f"madeup-{uuid.uuid4()}",
        version="1.2.3",
        channel_id=channel.id,
    )
    db.add(conda_package)
    db.commit()

    conda_package_build = orm.CondaPackageBuild(
        package_id=conda_package.id,
        build="fakebuild",
        build_number=1,
        constrains=[],
        depends=[],
        md5=str(uuid.uuid4()),
        sha256=str(uuid.uuid4()),
        size=123456,
        subdir="noarch",
        timestamp=12345667,
    )
    db.add(conda_package_build)
    db.commit()

    build.package_builds.append(conda_package_build)
    db.commit()


def _create_build_artifacts(db: Session, conda_store, build: orm.Build):
    conda_store.storage.set(
        db,
        build.id,
        build.log_key,
        b"fake logs",
        content_type="text/plain",
        artifact_type=schema.BuildArtifactType.LOGS,
    )

    directory_build_artifact = orm.BuildArtifact(
        build_id=build.id,
        artifact_type=schema.BuildArtifactType.DIRECTORY,
        key=str(build.build_path(conda_store)),
    )
    db.add(directory_build_artifact)

    lockfile_build_artifact = orm.BuildArtifact(
        build_id=build.id, artifact_type=schema.BuildArtifactType.LOCKFILE, key=""
    )
    db.add(lockfile_build_artifact)

    conda_store.storage.set(
        db,
        build.id,
        build.conda_env_export_key,
        json.dumps(
            dict(name="testing", channels=["conda-forge"], dependencies=["numpy"])
        ).encode("utf-8"),
        content_type="text/yaml",
        artifact_type=schema.BuildArtifactType.YAML,
    )

    conda_store.storage.set(
        db,
        build.id,
        build.conda_pack_key,
        b"testing-conda-package",
        content_type="application/gzip",
        artifact_type=schema.BuildArtifactType.CONDA_PACK,
    )

    # have not included docker at the moment



"""lockfile spec

Revision ID: bf065abf375b
Revises: e17b4cc6e086
Create Date: 2024-03-02 09:21:02.519805

"""

import sqlalchemy as sa

from alembic import op


# revision identifiers, used by Alembic.
revision = "bf065abf375b"
down_revision = "e17b4cc6e086"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "specification",
        # https://github.com/sqlalchemy/sqlalchemy/issues/1403#issuecomment-1698365595
        sa.Column(
            "is_lockfile", sa.Boolean(), nullable=False, server_default=sa.sql.false()
        ),
    )


def downgrade():
    op.drop_column("specification", "is_lockfile")



def logged_command(context, command, **kwargs):
    # This is here only for backward compatibility, new code should use the
    # run_command method instead of calling this function
    context.run_command(command, **kwargs)



import sys

from celery.result import AsyncResult

from conda_store_server import api, schema


def test_conda_store_app_register_solve(db, conda_store, celery_worker):
    conda_specification = schema.CondaSpecification(
        name="pytest-name",
        channels=["main"],
        dependencies=["python"],
    )

    task_id, solve_id = conda_store.register_solve(db, conda_specification)
    solve = api.get_solve(db, solve_id=solve_id)

    assert solve is not None
    assert solve.started_on is None
    assert solve.ended_on is None
    assert solve.package_builds == []
    assert solve.specification.spec["name"] == conda_specification.name
    assert solve.specification.spec["channels"] == conda_specification.channels
    assert solve.specification.spec["dependencies"] == conda_specification.dependencies
    task = AsyncResult(task_id)
    task.get(timeout=30)
    assert task.state == "SUCCESS"

    db.expire_all()
    assert solve.ended_on is not None
    assert len(solve.package_builds) > 0


def test_conda_store_register_environment_workflow(db, conda_store, celery_worker):
    """Test entire environment build workflow"""
    conda_specification = schema.CondaSpecification(
        name="pytest-name",
        channels=["main"],
        dependencies=["python"],
    )
    namespace_name = "pytest-namespace"

    build_id = conda_store.register_environment(
        db, specification=conda_specification.dict(), namespace=namespace_name
    )

    build = api.get_build(db, build_id=build_id)
    assert build is not None
    assert build.status == schema.BuildStatus.QUEUED
    assert build.environment.name == conda_specification.name
    assert build.environment.namespace.name == namespace_name
    assert build.specification.spec["name"] == conda_specification.name
    assert build.specification.spec["channels"] == conda_specification.channels
    assert build.specification.spec["dependencies"] == conda_specification.dependencies
    # when new environment is created current_build should be the default build
    assert build.environment.current_build == build

    from celery.result import AsyncResult

    # wait for task to complete
    # build: environment, export, archive
    # docker is expected to fail will be fixed soon

    task = AsyncResult(f"build-{build.id}-environment")
    task.wait(timeout=60)

    task = AsyncResult(f"build-{build.id}-conda-env-export")
    task.wait(timeout=60)

    task = AsyncResult(f"build-{build.id}-conda-pack")
    task.wait(timeout=60)

    if sys.platform == "linux":
        task = AsyncResult(f"build-{build.id}-docker")
        task.wait(timeout=2 * 60)

    task = AsyncResult(f"build-{build.id}-constructor-installer")
    task.wait(timeout=5 * 60)

    db.expire_all()
    assert build.status == schema.BuildStatus.COMPLETED


def test_conda_store_register_environment_force_false_same_namespace(db, conda_store):
    """Ensure behavior that when force=False and same namespace the
    same spec does not trigger another build

    """
    conda_specification = schema.CondaSpecification(
        name="pytest-name",
        channels=["main"],
        dependencies=["python"],
    )
    namespace_name = "pytest-namespace"

    first_build_id = conda_store.register_environment(
        db,
        specification=conda_specification.dict(),
        namespace=namespace_name,
        force=False,
    )

    second_build_id = conda_store.register_environment(
        db,
        specification=conda_specification.dict(),
        namespace=namespace_name,
        force=False,
    )

    assert first_build_id == 1
    assert second_build_id is None


def test_conda_store_register_environment_force_false_different_namespace(
    db, conda_store
):
    """Ensure behavior that when force=False and different namespace
    the same spec still triggers another build

    """
    conda_specification = schema.CondaSpecification(
        name="pytest-name",
        channels=["main"],
        dependencies=["python"],
    )

    first_build_id = conda_store.register_environment(
        db,
        specification=conda_specification.dict(),
        namespace="pytest-namespace",
        force=False,
    )

    second_build_id = conda_store.register_environment(
        db,
        specification=conda_specification.dict(),
        namespace="pytest-different-namespace",
        force=False,
    )

    assert first_build_id == 1
    assert second_build_id == 2


def test_conda_store_register_environment_duplicate_force_true(db, conda_store):
    """Ensure behavior that when force=True the same spec in same
    namespace still triggers another build

    """
    conda_specification = schema.CondaSpecification(
        name="pytest-name",
        channels=["main"],
        dependencies=["python"],
    )
    namespace_name = "pytest-namespace"

    first_build_id = conda_store.register_environment(
        db,
        specification=conda_specification.dict(),
        namespace=namespace_name,
        force=True,
    )

    second_build_id = conda_store.register_environment(
        db,
        specification=conda_specification.dict(),
        namespace=namespace_name,
        force=True,
    )

    assert first_build_id == 1
    assert second_build_id == 2



import datetime
import os
import sys

from contextlib import contextmanager
from typing import Any, Dict

import pydantic

from celery import Celery, group
from sqlalchemy.orm import Session
from sqlalchemy.pool import QueuePool
from traitlets import (
    Bool,
    Callable,
    Integer,
    List,
    TraitError,
    Type,
    Unicode,
    Union,
    default,
    validate,
)
from traitlets.config import LoggingConfigurable

from conda_store_server import (
    CONDA_STORE_DIR,
    BuildKey,
    api,
    conda_utils,
    environment,
    orm,
    registry,
    schema,
    storage,
    utils,
)


def conda_store_validate_specification(
    db: Session,
    conda_store: "CondaStore",
    namespace: str,
    specification: schema.CondaSpecification,
) -> schema.CondaSpecification:
    settings = conda_store.get_settings(
        db, namespace=namespace, environment_name=specification.name
    )

    specification = environment.validate_environment_channels(specification, settings)
    specification = environment.validate_environment_pypi_packages(
        specification, settings
    )
    specification = environment.validate_environment_conda_packages(
        specification, settings
    )

    return specification


def conda_store_validate_action(
    db: Session,
    conda_store: "CondaStore",
    namespace: str,
    action: schema.Permissions,
) -> None:
    settings = conda_store.get_settings(db)
    system_metrics = api.get_system_metrics(db)

    if action in (
        schema.Permissions.ENVIRONMENT_CREATE,
        schema.Permissions.ENVIRONMENT_UPDATE,
    ) and (settings.storage_threshold > system_metrics.disk_free):
        raise utils.CondaStoreError(
            f"`CondaStore.storage_threshold` reached. Action {action.value} prevented due to insufficient storage space"
        )


class CondaStore(LoggingConfigurable):
    storage_class = Type(
        default_value=storage.LocalStorage,
        klass=storage.Storage,
        allow_none=False,
        config=True,
    )

    container_registry_class = Type(
        default_value=registry.ContainerRegistry,
        klass=registry.ContainerRegistry,
        allow_none=False,
        config=True,
    )

    store_directory = Unicode(
        str(CONDA_STORE_DIR / "state"),
        help="directory for conda-store to build environments and store state",
        config=True,
    )

    build_directory = Unicode(
        "{store_directory}/{namespace}",
        help="Template used to form the directory for storing conda environment builds. Available keys: store_directory, namespace, name. The default will put all built environments in the same namespace within the same directory.",
        config=True,
    )

    environment_directory = Unicode(
        "{store_directory}/{namespace}/envs/{name}",
        help="Template used to form the directory for symlinking conda environment builds. Available keys: store_directory, namespace, name. The default will put all environments in the same namespace within the same directory.",
        config=True,
    )

    build_key_version = Integer(
        BuildKey.set_current_version(2),
        help="Build key version to use: 1 (long, legacy), 2 (shorter hash, default), 3 (hash-only, experimental)",
        config=True,
    )

    @validate("build_key_version")
    def _check_build_key_version(self, proposal):
        try:
            return BuildKey.set_current_version(proposal.value)
        except Exception as e:
            raise TraitError(f"c.CondaStore.build_key_version: {e}")

    win_extended_length_prefix = Bool(
        False,
        help="Use the extended-length prefix '\\\\?\\' (Windows-only), default: False",
        config=True,
    )

    conda_command = Unicode(
        "mamba",
        help="conda executable to use for solves",
        config=True,
    )

    conda_solve_platforms = List(
        [conda_utils.conda_platform()],
        description="Conda platforms to solve environments for via conda-lock. Must include current platform.",
        config=True,
    )

    conda_channel_alias = Unicode(
        "https://conda.anaconda.org",
        help="The prepended url location to associate with channel names",
        config=True,
    )

    conda_flags = Unicode(
        "--strict-channel-priority",
        help="The flags to be passed through the CONDA_FLAGS environment variable during the environment build",
        config=True,
    )

    conda_platforms = List(
        [conda_utils.conda_platform(), "noarch"],
        help="Conda platforms to download package repodata.json from. By default includes current architecture and noarch",
        config=True,
    )

    conda_default_channels = List(
        ["conda-forge"],
        help="Conda channels that by default are included if channels are empty",
        config=True,
    )

    conda_allowed_channels = List(
        [],
        help=(
            "Allowed conda channels to be used in conda environments. "
            "If set to empty list all channels are accepted (default). "
            "Example: "
            '["main", "conda-forge", "https://repo.anaconda.com/pkgs/main"]'
        ),
        config=True,
    )

    conda_indexed_channels = List(
        ["main", "conda-forge", "https://repo.anaconda.com/pkgs/main"],
        help="Conda channels to be indexed by conda-store at start.  Defaults to main and conda-forge.",
        config=True,
    )

    conda_default_packages = List(
        [],
        help="Conda packages that included by default if none are included",
        config=True,
    )

    conda_required_packages = List(
        [],
        help="Conda packages that are required to be within environment specification. Will raise a validation error is package not in specification",
        config=True,
    )

    conda_included_packages = List(
        [],
        help="Conda packages that auto included within environment specification. Will not raise a validation error if package not in specification and will be auto added",
        config=True,
    )

    pypi_default_packages = List(
        [],
        help="PyPi packages that included by default if none are included",
        config=True,
    )

    pypi_required_packages = List(
        [],
        help="PyPi packages that are required to be within environment specification. Will raise a validation error is package not in specification",
        config=True,
    )

    pypi_included_packages = List(
        [],
        help="PyPi packages that auto included within environment specification. Will not raise a validation error if package not in specification and will be auto added",
        config=True,
    )

    conda_max_solve_time = Integer(
        5 * 60,  # 5 minute
        help="Maximum time in seconds to allow for solving a given conda environment",
        config=True,
    )

    storage_threshold = Integer(
        5 * 1024**3,  # 5 GB
        help="Storage threshold in bytes of minimum available storage required in order to perform builds",
        config=True,
    )

    database_url = Unicode(
        "sqlite:///" + str(CONDA_STORE_DIR / "conda-store.sqlite"),
        help="url for the database. e.g. 'sqlite:///conda-store.sqlite' tables will be automatically created if they do not exist",
        config=True,
    )

    upgrade_db = Bool(
        True,
        help="""Upgrade the database automatically on start.
        Only safe if database is regularly backed up.
        """,
        config=True,
    )

    redis_url = Unicode(
        None,
        help="Redis connection url in form 'redis://:<password>@<hostname>:<port>/0'. Connection is used by Celery along with conda-store internally",
        config=True,
        allow_none=True,
    )

    @validate("redis_url")
    def _check_redis(self, proposal):
        try:
            if self.redis_url is not None:
                self.redis.ping()
        except Exception:
            raise TraitError(
                f'c.CondaStore.redis_url unable to connect with Redis database at "{self.redis_url}"'
            )
        return proposal.value

    celery_broker_url = Unicode(
        help="broker url to use for celery tasks",
        config=True,
    )

    build_artifacts = List(
        [
            schema.BuildArtifactType.LOCKFILE,
            schema.BuildArtifactType.YAML,
            schema.BuildArtifactType.CONDA_PACK,
            schema.BuildArtifactType.CONSTRUCTOR_INSTALLER,
            *(
                [
                    schema.BuildArtifactType.DOCKER_MANIFEST,
                    schema.BuildArtifactType.CONTAINER_REGISTRY,
                ]
                if sys.platform == "linux"
                else []
            ),
        ],
        help="artifacts to build in conda-store. By default all of the artifacts",
        config=True,
    )

    build_artifacts_kept_on_deletion = List(
        [
            schema.BuildArtifactType.LOGS,
            schema.BuildArtifactType.LOCKFILE,
            schema.BuildArtifactType.YAML,
            # no possible way to delete these artifacts
            # in most container registries via api
            schema.BuildArtifactType.CONTAINER_REGISTRY,
        ],
        help="artifacts to keep on build deletion",
        config=True,
    )

    serialize_builds = Bool(
        True,
        help="DEPRICATED no longer has any effect",
        config=True,
    )

    @default("celery_broker_url")
    def _default_celery_broker_url(self):
        if self.redis_url is not None:
            return self.redis_url
        return f"sqla+{self.database_url}"

    celery_results_backend = Unicode(
        help="backend to use for celery task results",
        config=True,
    )

    @default("celery_results_backend")
    def _default_celery_results_backend(self):
        if self.redis_url is not None:
            return self.redis_url
        return f"db+{self.database_url}"

    default_namespace = Unicode(
        "default", help="default namespace for conda-store", config=True
    )

    filesystem_namespace = Unicode(
        "filesystem",
        help="namespace to use for environments picked up via `CondaStoreWorker.watch_paths` on the filesystem",
        config=True,
    )

    default_uid = Integer(
        None if sys.platform == "win32" else os.getuid(),
        help="default uid to assign to built environments",
        config=True,
        allow_none=True,
    )

    default_gid = Integer(
        None if sys.platform == "win32" else os.getgid(),
        help="default gid to assign to built environments",
        config=True,
        allow_none=True,
    )

    default_permissions = Unicode(
        None if sys.platform == "win32" else "775",
        help="default file permissions to assign to built environments",
        config=True,
        allow_none=True,
    )

    default_docker_base_image = Union(
        [Unicode(), Callable()],
        help="default base image used for the Dockerized environments. Make sure to have a proper glibc within image (highly discourage alpine/musl based images). Can also be callable function which takes the `orm.Build` object as input which has access to all attributes about the build such as install packages, requested packages, name, namespace, etc",
        config=True,
    )

    @default("default_docker_base_image")
    def _default_docker_base_image(self):
        def _docker_base_image(build: orm.Build):
            return "registry-1.docker.io/library/debian:sid-slim"

        return _docker_base_image

    validate_specification = Callable(
        conda_store_validate_specification,
        help="callable function taking conda_store, namespace, and specification as input arguments to apply for validating and modifying a given specification. If there are validation issues with the environment ValueError with message should be raised. If changed you may need to call the default function to preseve many of the trait effects e.g. `c.CondaStore.default_channels` etc",
        config=True,
    )

    validate_action = Callable(
        conda_store_validate_action,
        help="callable function taking conda_store, namespace, and action. If there are issues with performing the given action raise a CondaStoreError should be raised.",
        config=True,
    )

    post_update_environment_build_hook = Callable(
        default_value=None,
        help="callable function taking conda_store and `orm.Environment` object as input arguments. This function can be used to add custom behavior that will run after an environment's current build changes.",
        config=True,
        allow_none=True,
    )

    @property
    def session_factory(self):
        if hasattr(self, "_session_factory"):
            return self._session_factory

        self._session_factory = orm.new_session_factory(
            url=self.database_url,
            poolclass=QueuePool,
        )

        return self._session_factory

    # Do not define this as a FastAPI dependency! That would cause Sessions
    # not to be closed, which would lead to DB pool exhaustion and requests
    # getting blocked.
    # https://github.com/conda-incubator/conda-store/issues/598
    @contextmanager
    def get_db(self):
        db = self.session_factory()
        try:
            yield db
        finally:
            db.close()

    @property
    def redis(self):
        import redis

        if hasattr(self, "_redis"):
            return self._redis
        self._redis = redis.Redis.from_url(self.redis_url)
        return self._redis

    def configuration(self, db: Session):
        return orm.CondaStoreConfiguration.configuration(db)

    @property
    def storage(self):
        if hasattr(self, "_storage"):
            return self._storage
        self._storage = self.storage_class(parent=self, log=self.log)

        if isinstance(self._storage, storage.LocalStorage):
            os.makedirs(self._storage.storage_path, exist_ok=True)

        return self._storage

    @property
    def container_registry(self):
        if hasattr(self, "_container_registry"):
            return self._container_registry
        self._container_registry = self.container_registry_class(
            parent=self, log=self.log
        )
        return self._container_registry

    @property
    def celery_config(self):
        return {
            "broker_url": self.celery_broker_url,
            "result_backend": self.celery_results_backend,
            "imports": [
                "conda_store_server.worker.tasks",
                "celery.contrib.testing.tasks",
            ],
            "task_track_started": True,
            "result_extended": True,
            "beat_schedule": {
                "watch-paths": {
                    "task": "task_watch_paths",
                    "schedule": 60.0,  # 1 minute
                    "args": [],
                    "kwargs": {},
                },
                "update-conda-channels": {
                    "task": "task_update_conda_channels",
                    "schedule": 15.0 * 60.0,  # 15 minutes
                    "args": [],
                    "kwargs": {},
                },
            },
            "beat_schedule_filename": str(CONDA_STORE_DIR / "celerybeat-schedule"),
            "triatlets": {},
        }

    @property
    def celery_app(self):
        # if hasattr(self, "_celery_app"):
        #     return self._celery_app

        self._celery_app = Celery("tasks")
        self._celery_app.config_from_object(self.celery_config)
        return self._celery_app

    def ensure_settings(self, db: Session):
        """Ensure that conda-store traitlets settings are applied"""
        settings = schema.Settings(
            default_namespace=self.default_namespace,
            filesystem_namespace=self.filesystem_namespace,
            default_uid=self.default_uid,
            default_gid=self.default_gid,
            default_permissions=self.default_permissions,
            storage_threshold=self.storage_threshold,
            conda_command=self.conda_command,
            conda_platforms=self.conda_platforms,
            conda_max_solve_time=self.conda_max_solve_time,
            conda_indexed_channels=self.conda_indexed_channels,
            build_artifacts_kept_on_deletion=self.build_artifacts_kept_on_deletion,
            conda_solve_platforms=self.conda_solve_platforms,
            conda_channel_alias=self.conda_channel_alias,
            conda_default_channels=self.conda_default_channels,
            conda_allowed_channels=self.conda_allowed_channels,
            conda_default_packages=self.conda_default_packages,
            conda_required_packages=self.conda_required_packages,
            conda_included_packages=self.conda_included_packages,
            pypi_default_packages=self.pypi_default_packages,
            pypi_required_packages=self.pypi_required_packages,
            pypi_included_packages=self.pypi_included_packages,
            build_artifacts=self.build_artifacts,
            # default_docker_base_image=self.default_docker_base_image,
        )
        api.set_kvstore_key_values(db, "setting", settings.dict(), update=False)

    def ensure_namespace(self, db: Session):
        """Ensure that conda-store default namespaces exists"""
        api.ensure_namespace(db, self.default_namespace)

    def ensure_directories(self):
        """Ensure that conda-store filesystem directories exist"""
        os.makedirs(self.store_directory, exist_ok=True)

    def ensure_conda_channels(self, db: Session):
        """Ensure that conda-store indexed channels and packages are in database"""
        self.log.info("updating conda store channels")

        settings = self.get_settings(db)

        for channel in settings.conda_indexed_channels:
            normalized_channel = conda_utils.normalize_channel_name(
                settings.conda_channel_alias, channel
            )
            api.ensure_conda_channel(db, normalized_channel)

    def set_settings(
        self,
        db: Session,
        namespace: str = None,
        environment_name: str = None,
        data: Dict[str, Any] = {},
    ):
        setting_keys = schema.Settings.__fields__.keys()
        if not data.keys() <= setting_keys:
            invalid_keys = data.keys() - setting_keys
            raise ValueError(f"Invalid setting keys {invalid_keys}")

        for key, value in data.items():
            field = schema.Settings.__fields__[key]
            global_setting = field.field_info.extra["metadata"]["global"]
            if global_setting and (
                namespace is not None or environment_name is not None
            ):
                raise ValueError(
                    f"Setting {key} is a global setting cannot be set within namespace or environment"
                )

            try:
                pydantic.parse_obj_as(field.outer_type_, value)
            except Exception as e:
                raise ValueError(
                    f"Invalid parsing of setting {key} expected type {field.outer_type_} ran into error {e}"
                )

        if namespace is not None and environment_name is not None:
            prefix = f"setting/{namespace}/{environment_name}"
        elif namespace is not None:
            prefix = f"setting/{namespace}"
        else:
            prefix = "setting"

        api.set_kvstore_key_values(db, prefix, data)

    def get_settings(
        self, db: Session, namespace: str = None, environment_name: str = None
    ) -> schema.Settings:
        # setting logic is intentionally done in python code
        # rather than using the database for merges and ordering
        # becuase in the future we may likely want to do some
        # more complex logic around settings

        prefixes = ["setting"]
        if namespace is not None:
            prefixes.append(f"setting/{namespace}")
        if namespace is not None and environment_name is not None:
            prefixes.append(f"setting/{namespace}/{environment_name}")

        settings = {}
        for prefix in prefixes:
            settings.update(api.get_kvstore_key_values(db, prefix))

        return schema.Settings(**settings)

    def register_solve(self, db: Session, specification: schema.CondaSpecification):
        """Registers a solve for a given specification"""
        settings = self.get_settings(db)

        self.validate_action(
            db=db,
            conda_store=self,
            namespace="solve",
            action=schema.Permissions.ENVIRONMENT_SOLVE,
        )

        specification_model = self.validate_specification(
            db=db,
            conda_store=self,
            namespace="solve",
            specification=specification,
        )

        specification_orm = api.ensure_specification(db, specification_model)
        solve = api.create_solve(db, specification_orm.id)
        db.commit()

        self.celery_app

        from conda_store_server.worker import tasks

        task_id = f"solve-{solve.id}"
        tasks.task_solve_conda_environment.apply_async(
            args=[solve.id],
            time_limit=settings.conda_max_solve_time,
            task_id=task_id,
        )

        return task_id, solve.id

    def register_environment(
        self,
        db: Session,
        specification: dict,
        namespace: str = None,
        force: bool = True,
        is_lockfile: bool = False,
    ):
        """Register a given specification to conda store with given namespace/name."""

        settings = self.get_settings(db)

        namespace = namespace or settings.default_namespace
        namespace = api.ensure_namespace(db, name=namespace)

        self.validate_action(
            db=db,
            conda_store=self,
            namespace=namespace.name,
            action=schema.Permissions.ENVIRONMENT_CREATE,
        )

        if is_lockfile:
            # It's a lockfile, do not do any validation in this case. If there
            # are problems, these would be caught earlier during parsing or
            # later when conda-lock attempts to install it.
            specification_model = specification
        else:
            specification_model = self.validate_specification(
                db=db,
                conda_store=self,
                namespace=namespace.name,
                specification=schema.CondaSpecification.parse_obj(specification),
            )

        spec_sha256 = utils.datastructure_hash(specification_model.dict())
        matching_specification = api.get_specification(db, sha256=spec_sha256)
        if (
            matching_specification is not None
            and not force
            and any(
                _.environment.namespace.id == namespace.id
                for _ in matching_specification.builds
            )
        ):
            return None

        specification = api.ensure_specification(
            db, specification_model, is_lockfile=is_lockfile
        )
        environment_was_empty = (
            api.get_environment(db, name=specification.name, namespace_id=namespace.id)
            is None
        )
        environment = api.ensure_environment(
            db,
            name=specification.name,
            namespace_id=namespace.id,
            description=specification.spec["description"],
        )

        build = self.create_build(db, environment.id, specification.sha256)

        if environment_was_empty:
            environment.current_build = build
            db.commit()

        return build.id

    def create_build(self, db: Session, environment_id: int, specification_sha256: str):
        environment = api.get_environment(db, id=environment_id)
        self.validate_action(
            db=db,
            conda_store=self,
            namespace=environment.namespace.name,
            action=schema.Permissions.ENVIRONMENT_UPDATE,
        )

        settings = self.get_settings(
            db, namespace=environment.namespace.name, environment_name=environment.name
        )

        specification = api.get_specification(db, specification_sha256)
        build = api.create_build(
            db, environment_id=environment_id, specification_id=specification.id
        )
        db.commit()

        self.celery_app

        # must import tasks after a celery app has been initialized
        from conda_store_server.worker import tasks

        # Note: task ids used here must also be in api_put_build_cancel

        artifact_tasks = []
        if schema.BuildArtifactType.YAML in settings.build_artifacts:
            artifact_tasks.append(
                tasks.task_build_conda_env_export.subtask(
                    args=(build.id,),
                    task_id=f"build-{build.id}-conda-env-export",
                    immutable=True,
                )
            )
        if schema.BuildArtifactType.CONDA_PACK in settings.build_artifacts:
            artifact_tasks.append(
                tasks.task_build_conda_pack.subtask(
                    args=(build.id,),
                    task_id=f"build-{build.id}-conda-pack",
                    immutable=True,
                )
            )

        if (
            schema.BuildArtifactType.DOCKER_MANIFEST in settings.build_artifacts
            or schema.BuildArtifactType.CONTAINER_REGISTRY in settings.build_artifacts
        ):
            artifact_tasks.append(
                tasks.task_build_conda_docker.subtask(
                    args=(build.id,), task_id=f"build-{build.id}-docker", immutable=True
                )
            )

        if schema.BuildArtifactType.CONSTRUCTOR_INSTALLER in settings.build_artifacts:
            artifact_tasks.append(
                tasks.task_build_constructor_installer.subtask(
                    args=(build.id,),
                    task_id=f"build-{build.id}-constructor-installer",
                    immutable=True,
                )
            )

        (
            tasks.task_update_storage_metrics.si()
            | tasks.task_build_conda_environment.subtask(
                args=(build.id,),
                task_id=f"build-{build.id}-environment",
                immutable=True,
            )
            | group(*artifact_tasks)
            | tasks.task_update_storage_metrics.si()
        ).apply_async(task_id=f"build-{build.id}")

        return build

    def update_environment_build(
        self, db: Session, namespace: str, name: str, build_id: int
    ):
        self.validate_action(
            db=db,
            conda_store=self,
            namespace=namespace,
            action=schema.Permissions.ENVIRONMENT_UPDATE,
        )

        build = api.get_build(db, build_id)
        if build is None:
            raise utils.CondaStoreError(f"build id={build_id} does not exist")

        environment = api.get_environment(db, namespace=namespace, name=name)
        if environment is None:
            raise utils.CondaStoreError(
                f"environment namespace={namespace} name={name} does not exist"
            )

        if build.status != schema.BuildStatus.COMPLETED:
            raise utils.CondaStoreError(
                "cannot update environment to build id since not completed"
            )

        if build.specification.name != name:
            raise utils.CondaStoreError(
                "cannot update environment to build id since specification does not match environment name"
            )

        environment.current_build_id = build.id
        db.commit()

        self.celery_app
        # must import tasks after a celery app has been initialized
        from conda_store_server.worker import tasks

        tasks.task_update_environment_build.si(environment.id).apply_async()

    def update_environment_description(
        self, db: Session, namespace: str, name: str, description: str
    ):
        environment = api.get_environment(db, namespace=namespace, name=name)
        if environment is None:
            raise utils.CondaStoreError(
                f"environment namespace={namespace} name={name} does not exist"
            )

        environment.description = description
        db.commit()

    def delete_namespace(self, db: Session, namespace: str):
        self.validate_action(
            db=db,
            conda_store=self,
            namespace=namespace,
            action=schema.Permissions.NAMESPACE_DELETE,
        )

        namespace = api.get_namespace(db, name=namespace)
        if namespace is None:
            raise utils.CondaStoreError(f"namespace={namespace} does not exist")

        utcnow = datetime.datetime.utcnow()
        namespace.deleted_on = utcnow
        for environment_orm in namespace.environments:
            environment_orm.deleted_on = utcnow
            for build in environment_orm.builds:
                build.deleted_on = utcnow
        db.commit()

        self.celery_app

        # must import tasks after a celery app has been initialized
        from conda_store_server.worker import tasks

        tasks.task_delete_namespace.si(namespace.id).apply_async()

    def delete_environment(self, db: Session, namespace: str, name: str):
        self.validate_action(
            db=db,
            conda_store=self,
            namespace=namespace,
            action=schema.Permissions.ENVIRONMENT_DELETE,
        )

        environment = api.get_environment(db, namespace=namespace, name=name)
        if environment is None:
            raise utils.CondaStoreError(
                f"environment namespace={namespace} name={name} does not exist"
            )

        utcnow = datetime.datetime.utcnow()
        environment.deleted_on = utcnow
        for build in environment.builds:
            build.deleted_on = utcnow
        db.commit()

        self.celery_app

        # must import tasks after a celery app has been initialized
        from conda_store_server.worker import tasks

        tasks.task_delete_environment.si(environment.id).apply_async()

    def delete_build(self, db: Session, build_id: int):
        build = api.get_build(db, build_id)

        self.validate_action(
            db=db,
            conda_store=self,
            namespace=build.environment.namespace.name,
            action=schema.Permissions.BUILD_DELETE,
        )

        if build.status not in [
            schema.BuildStatus.FAILED,
            schema.BuildStatus.COMPLETED,
        ]:
            raise utils.CondaStoreError(
                "cannot delete build since not finished building"
            )

        build.deleted_on = datetime.datetime.utcnow()
        db.commit()

        self.celery_app

        # must import tasks after a celery app has been initialized
        from conda_store_server.worker import tasks

        tasks.task_delete_build.si(build.id).apply_async()



import datetime
import pathlib
import sys

import pytest
import yaml

from fastapi.testclient import TestClient


from conda_store_server import (  # isort:skip
    action,
    api,
    app,
    dbutil,
    schema,
    storage,
    testing,
    utils,
)

from conda_store_server.server import app as server_app  # isort:skip


@pytest.fixture
def celery_config(tmp_path, conda_store):
    config = conda_store.celery_config
    config["traitlets"] = {
        "CondaStore": {
            "database_url": conda_store.database_url,
            "store_directory": conda_store.store_directory,
        }
    }
    config["beat_schedule_filename"] = str(
        tmp_path / ".conda-store" / "celerybeat-schedule"
    )
    return config


@pytest.fixture
def conda_store_config(tmp_path, request):
    from traitlets.config import Config

    filename = tmp_path / ".conda-store" / "database.sqlite"

    store_directory = tmp_path / ".conda-store" / "state"
    store_directory.mkdir(parents=True)

    storage.LocalStorage.storage_path = str(tmp_path / ".conda-store" / "storage")

    with utils.chdir(tmp_path):
        yield Config(
            CondaStore=dict(
                storage_class=storage.LocalStorage,
                store_directory=str(store_directory),
                database_url=f"sqlite:///{filename}?check_same_thread=False",
            )
        )

    original_sys_argv = list(sys.argv)
    sys.argv = [sys.argv[0]]

    def teardown():
        sys.argv = list(original_sys_argv)

    request.addfinalizer(teardown)


@pytest.fixture
def conda_store_server(conda_store_config):
    _conda_store_server = server_app.CondaStoreServer(config=conda_store_config)
    _conda_store_server.initialize()

    _conda_store = _conda_store_server.conda_store

    pathlib.Path(_conda_store.store_directory).mkdir(exist_ok=True)

    dbutil.upgrade(_conda_store.database_url)

    with _conda_store.session_factory() as db:
        _conda_store.ensure_settings(db)
        _conda_store.configuration(db).update_storage_metrics(
            db, _conda_store.store_directory
        )

        _conda_store.celery_app

        # must import tasks after a celery app has been initialized
        # ensure that models are created
        from celery.backends.database.session import ResultModelBase

        import conda_store_server.worker.tasks  # noqa

        ResultModelBase.metadata.create_all(db.get_bind())

    yield _conda_store_server


@pytest.fixture
def testclient(conda_store_server):
    return TestClient(conda_store_server.init_fastapi_app())


@pytest.fixture
def authenticate(testclient):
    response = testclient.post(
        "/login/", json={"username": "username", "password": "password"}
    )
    assert response.status_code == 200


@pytest.fixture
def seed_conda_store(db, conda_store):
    testing.seed_conda_store(
        db,
        conda_store,
        {
            "default": {
                "name1": schema.CondaSpecification(
                    name="name1",
                    channels=["conda-forge"],
                    dependencies=["numpy"],
                ),
                "name2": schema.CondaSpecification(
                    name="name2",
                    channels=["defaults"],
                    dependencies=["flask"],
                ),
            },
            "namespace1": {
                "name3": schema.CondaSpecification(
                    name="name3",
                    channels=["bioconda"],
                    dependencies=["numba"],
                )
            },
            "namespace2": {
                "name4": schema.CondaSpecification(
                    name="name4",
                    channels=["bioconda"],
                    dependencies=["numba"],
                )
            },
        },
    )

    # for testing purposes make build 4 complete
    build = api.get_build(db, build_id=4)
    build.started_on = datetime.datetime.utcnow()
    build.ended_on = datetime.datetime.utcnow()
    build.status = schema.BuildStatus.COMPLETED
    db.commit()


@pytest.fixture
def conda_store(conda_store_config):
    _conda_store = app.CondaStore(config=conda_store_config)

    pathlib.Path(_conda_store.store_directory).mkdir(exist_ok=True)

    dbutil.upgrade(_conda_store.database_url)

    with _conda_store.session_factory() as db:
        _conda_store.ensure_settings(db)
        _conda_store.configuration(db).update_storage_metrics(
            db, _conda_store.store_directory
        )

        _conda_store.celery_app

        # must import tasks after a celery app has been initialized
        # ensure that models are created
        from celery.backends.database.session import ResultModelBase

        import conda_store_server.worker.tasks  # noqa

        ResultModelBase.metadata.create_all(db.get_bind())

    yield _conda_store


@pytest.fixture
def db(conda_store):
    with conda_store.session_factory() as _db:
        yield _db


@pytest.fixture
def simple_specification():
    yield schema.CondaSpecification(
        name="test",
        channels=["main"],
        dependencies=["zlib"],
    )


@pytest.fixture
def simple_specification_with_pip():
    yield schema.CondaSpecification(
        name="test",
        channels=["main"],
        dependencies=[
            "python",
            {"pip": ["flask"]},
        ],
    )


@pytest.fixture
def simple_conda_lock():
    with (pathlib.Path(__file__).parent / "assets/conda-lock.zlib.yaml").open() as f:
        return yaml.safe_load(f)


@pytest.fixture
def simple_conda_lock_with_pip():
    with (
        pathlib.Path(__file__).parent / "assets/conda-lock.zlib.flask.yaml"
    ).open() as f:
        return yaml.safe_load(f)


@pytest.fixture
def simple_lockfile_specification(simple_conda_lock):
    yield schema.LockfileSpecification.parse_obj(
        {
            "name": "test",
            "description": "simple lockfile specification",
            "lockfile": simple_conda_lock,
        }
    )


@pytest.fixture
def simple_lockfile_specification_with_pip(simple_conda_lock_with_pip):
    yield schema.LockfileSpecification.parse_obj(
        {
            "name": "test",
            "description": "simple lockfile specification with pip",
            "lockfile": simple_conda_lock_with_pip,
        }
    )


@pytest.fixture(
    params=[
        dict(
            name="test-prefix",
            channels=["main"],
            dependencies=["yaml"],
        ),
        dict(
            name="test-prefix",
            channels=["main"],
            dependencies=["python", {"pip": ["flask"]}],
        ),
    ]
)
def conda_prefix(conda_store, tmp_path, request):
    conda_prefix = tmp_path / "test-prefix"
    conda_prefix.mkdir()

    specification = schema.CondaSpecification(**request.param)

    action.action_install_specification(
        conda_command=conda_store.conda_command,
        specification=specification,
        conda_prefix=conda_prefix,
    )
    yield conda_prefix



import contextlib
import functools
import io
import logging
import subprocess
import tempfile
import typing
import uuid

from conda_store_server import utils


def action(f: typing.Callable):
    @functools.wraps(f)
    def wrapper(*args, stdout=None, stderr=None, **kwargs):
        action_context = ActionContext(stdout=stdout, stderr=stderr)
        with contextlib.ExitStack() as stack:
            # redirect stdout -> action_context.stdout
            stack.enter_context(contextlib.redirect_stdout(action_context.stdout))

            # redirect stderr -> action_context.stdout
            stack.enter_context(contextlib.redirect_stderr(action_context.stdout))

            # create a temporary directory
            tmpdir = stack.enter_context(tempfile.TemporaryDirectory())

            # enter temporary directory
            stack.enter_context(utils.chdir(tmpdir))

            # run function and store result
            action_context.result = f(action_context, *args, **kwargs)
        return action_context

    return wrapper


class ActionContext:
    def __init__(self, stdout=None, stderr=None):
        if stdout is not None and stderr is None:
            stderr = stdout

        self.id = str(uuid.uuid4())
        self.stdout = stdout if stdout is not None else io.StringIO()
        self.stderr = stderr if stderr is not None else io.StringIO()
        self.log = logging.getLogger(f"conda_store_server.action.{self.id}")
        self.log.propagate = False
        self.log.addHandler(logging.StreamHandler(stream=self.stdout))
        self.log.setLevel(logging.INFO)
        self.result = None
        self.artifacts = {}

    def run_command(self, command, redirect_stderr=True, **kwargs):
        """Runs command and immediately writes to logs"""
        self.log.info(f"Running command: {' '.join(command)}")

        # Unlike subprocess.run, Popen doesn't support the check argument, so
        # ignore it. The code below always checks the return code
        kwargs.pop("check", None)

        # https://stackoverflow.com/questions/4417546/constantly-print-subprocess-output-while-process-is-running
        with subprocess.Popen(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT if redirect_stderr else subprocess.PIPE,
            bufsize=1,
            universal_newlines=True,
            **kwargs,
        ) as p:
            for line in p.stdout:
                self.stdout.write(line)
            if not redirect_stderr:
                for line in p.stderr:
                    self.stderr.write(line)

        if p.returncode != 0:
            raise subprocess.CalledProcessError(p.returncode, p.args)

    def run(self, *args, redirect_stderr=True, **kwargs):
        """Runs command waiting for it to succeed before writing to logs"""
        result = subprocess.run(
            *args,
            **kwargs,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT if redirect_stderr else subprocess.PIPE,
            encoding="utf-8",
        )
        self.stdout.write(result.stdout)
        if not redirect_stderr:
            self.stderr.write(result.stderr)

        return result



import json
import pathlib

from conda_store_server import action


@action.action
def action_generate_conda_export(
    context, conda_command: str, conda_prefix: pathlib.Path
):
    command = [
        conda_command,
        "env",
        "export",
        "--prefix",
        str(conda_prefix),
        "--json",
    ]

    result = context.run(command, check=True, redirect_stderr=False)
    if result.stderr:
        context.log.warning(f"conda env export stderr: {result.stderr}")
    return json.loads(result.stdout)



import json
import os
import sys
import time

import pytest
import traitlets
import yaml

from conda_store_server import CONDA_STORE_DIR, __version__, schema


def test_api_version_unauth(testclient):
    response = testclient.get("/api/v1/")
    response.raise_for_status()

    r = schema.APIGetStatus.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK
    assert r.data.version == __version__


def test_api_permissions_unauth(testclient):
    response = testclient.get("api/v1/permission/")
    response.raise_for_status()

    r = schema.APIGetPermission.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK
    assert r.data.authenticated is False
    assert r.data.primary_namespace == "default"
    assert r.data.entity_permissions == {
        "default/*": sorted(
            [
                schema.Permissions.ENVIRONMENT_READ.value,
                schema.Permissions.NAMESPACE_READ.value,
                schema.Permissions.NAMESPACE_ROLE_MAPPING_READ.value,
            ]
        )
    }


def test_api_permissions_auth(testclient, authenticate):
    response = testclient.get("api/v1/permission/")
    response.raise_for_status()

    r = schema.APIGetPermission.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK
    assert r.data.authenticated is True
    assert r.data.primary_namespace == "username"
    assert r.data.entity_permissions == {
        "*/*": sorted(
            [
                schema.Permissions.ENVIRONMENT_CREATE.value,
                schema.Permissions.ENVIRONMENT_READ.value,
                schema.Permissions.ENVIRONMENT_UPDATE.value,
                schema.Permissions.ENVIRONMENT_DELETE.value,
                schema.Permissions.ENVIRONMENT_SOLVE.value,
                schema.Permissions.BUILD_CANCEL.value,
                schema.Permissions.BUILD_DELETE.value,
                schema.Permissions.NAMESPACE_CREATE.value,
                schema.Permissions.NAMESPACE_READ.value,
                schema.Permissions.NAMESPACE_DELETE.value,
                schema.Permissions.NAMESPACE_UPDATE.value,
                schema.Permissions.NAMESPACE_ROLE_MAPPING_CREATE.value,
                schema.Permissions.NAMESPACE_ROLE_MAPPING_READ.value,
                schema.Permissions.NAMESPACE_ROLE_MAPPING_UPDATE.value,
                schema.Permissions.NAMESPACE_ROLE_MAPPING_DELETE.value,
                schema.Permissions.SETTING_READ.value,
                schema.Permissions.SETTING_UPDATE.value,
            ]
        ),
        "default/*": sorted(
            [
                schema.Permissions.ENVIRONMENT_READ.value,
                schema.Permissions.NAMESPACE_READ.value,
                schema.Permissions.NAMESPACE_ROLE_MAPPING_READ.value,
            ]
        ),
        "filesystem/*": sorted(
            [
                schema.Permissions.ENVIRONMENT_READ.value,
                schema.Permissions.NAMESPACE_READ.value,
                schema.Permissions.NAMESPACE_ROLE_MAPPING_READ.value,
            ]
        ),
    }


def test_get_usage_unauth(testclient):
    response = testclient.get("/api/v1/usage/")
    response.raise_for_status()

    r = schema.APIPutSetting.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK


def test_get_usage_auth(testclient, authenticate):
    response = testclient.get("/api/v1/usage/")
    response.raise_for_status()

    r = schema.APIPutSetting.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK


def test_api_list_namespace_unauth(testclient, seed_conda_store):
    response = testclient.get("api/v1/namespace")
    response.raise_for_status()

    r = schema.APIListNamespace.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK
    # by default unauth only has access to the default namespace
    assert len(r.data) == 1
    assert r.data[0].name == "default"


def test_api_list_namespace_auth(testclient, seed_conda_store, authenticate):
    response = testclient.get("api/v1/namespace")
    response.raise_for_status()

    r = schema.APIListNamespace.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK
    assert sorted([_.name for _ in r.data]) == ["default", "namespace1", "namespace2"]


def test_api_get_namespace_unauth(testclient, seed_conda_store):
    response = testclient.get("api/v1/namespace/default")
    response.raise_for_status()

    r = schema.APIGetNamespace.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK
    assert r.data.name == "default"


def test_api_get_namespace_unauth_no_exist(testclient, seed_conda_store):
    response = testclient.get("api/v1/namespace/namespace1")
    assert response.status_code == 403

    r = schema.APIResponse.parse_obj(response.json())
    assert r.status == schema.APIStatus.ERROR


def test_api_get_namespace_auth(testclient, seed_conda_store, authenticate):
    response = testclient.get("api/v1/namespace/namespace1")
    response.raise_for_status()

    r = schema.APIGetNamespace.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK
    assert r.data.name == "namespace1"


def test_api_get_namespace_auth_no_exist(testclient, seed_conda_store, authenticate):
    response = testclient.get("api/v1/namespace/wrong")

    assert response.status_code == 404
    r = schema.APIResponse.parse_obj(response.json())
    assert r.status == schema.APIStatus.ERROR


def test_api_list_environments_unauth(testclient, seed_conda_store):
    response = testclient.get("api/v1/environment")
    response.raise_for_status()

    r = schema.APIListEnvironment.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK
    assert sorted([_.name for _ in r.data]) == ["name1", "name2"]


def test_api_list_environments_auth(testclient, seed_conda_store, authenticate):
    response = testclient.get("api/v1/environment")
    response.raise_for_status()

    r = schema.APIListEnvironment.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK
    assert sorted([_.name for _ in r.data]) == ["name1", "name2", "name3", "name4"]


def test_api_get_environment_unauth(testclient, seed_conda_store):
    response = testclient.get("api/v1/environment/namespace1/name3")
    assert response.status_code == 403

    r = schema.APIResponse.parse_obj(response.json())
    assert r.status == schema.APIStatus.ERROR


def test_api_get_environment_auth_existing(testclient, seed_conda_store, authenticate):
    response = testclient.get("api/v1/environment/namespace1/name3")
    response.raise_for_status()

    r = schema.APIGetEnvironment.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK
    assert r.data.namespace.name == "namespace1"
    assert r.data.name == "name3"


def test_api_get_environment_auth_not_existing(
    testclient, seed_conda_store, authenticate
):
    response = testclient.get("api/v1/environment/namespace1/wrong")
    assert response.status_code == 404

    r = schema.APIResponse.parse_obj(response.json())
    assert r.status == schema.APIStatus.ERROR


def test_api_list_builds_unauth(testclient, seed_conda_store):
    response = testclient.get("api/v1/build")
    response.raise_for_status()

    r = schema.APIListBuild.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK
    assert len(r.data) == 2


def test_api_list_builds_auth(testclient, seed_conda_store, authenticate):
    response = testclient.get("api/v1/build")
    response.raise_for_status()

    r = schema.APIListBuild.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK
    assert len(r.data) == 4


def test_api_get_build_one_unauth(testclient, seed_conda_store):
    response = testclient.get("api/v1/build/3")  # namespace1/name3
    assert response.status_code == 403

    r = schema.APIResponse.parse_obj(response.json())
    assert r.status == schema.APIStatus.ERROR


def test_api_get_build_one_auth(testclient, seed_conda_store, authenticate):
    response = testclient.get("api/v1/build/3")
    response.raise_for_status()

    r = schema.APIGetBuild.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK
    assert r.data.id == 3
    assert r.data.specification.name == "name3"
    assert r.data.status == schema.BuildStatus.QUEUED.value


def test_api_get_build_one_unauth_packages(testclient, seed_conda_store):
    response = testclient.get("api/v1/build/3/packages")
    assert response.status_code == 403

    r = schema.APIResponse.parse_obj(response.json())
    assert r.status == schema.APIStatus.ERROR


def test_api_get_build_one_auth_packages(testclient, seed_conda_store, authenticate):
    response = testclient.get("api/v1/build/3/packages")
    response.raise_for_status()

    r = schema.APIListCondaPackage.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK
    assert len(r.data) == 1


def test_api_get_build_auth_packages_no_exist(
    testclient, seed_conda_store, authenticate
):
    response = testclient.get("api/v1/build/101010101/packages")
    assert response.status_code == 404

    r = schema.APIResponse.parse_obj(response.json())
    assert r.status == schema.APIStatus.ERROR


def test_api_get_build_one_unauth_logs(testclient, seed_conda_store):
    response = testclient.get("api/v1/build/3/logs")
    assert response.status_code == 403

    r = schema.APIResponse.parse_obj(response.json())
    assert r.status == schema.APIStatus.ERROR


def test_api_get_build_one_auth_logs(testclient, seed_conda_store, authenticate):
    response = testclient.get("api/v1/build/3/logs")
    response.raise_for_status()

    logs = response.content.decode("utf-8")
    assert "fake logs" in logs


def test_api_get_build_auth_logs_no_exist(testclient, seed_conda_store, authenticate):
    response = testclient.get("api/v1/build/10101010101/logs")
    assert response.status_code == 404


def test_api_get_build_one_unauth_yaml(testclient, seed_conda_store):
    response = testclient.get("api/v1/build/3/yaml/")
    assert response.status_code == 403


def test_api_get_build_one_auth_yaml(testclient, seed_conda_store, authenticate):
    response = testclient.get("api/v1/build/3/yaml/")
    response.raise_for_status()

    environment_yaml = yaml.safe_load(response.content.decode("utf-8"))
    assert {"name", "channels", "dependencies"} <= environment_yaml.keys()


def test_api_get_build_two_auth(testclient, seed_conda_store, authenticate):
    response = testclient.get("api/v1/build/1010101010101")
    response.status_code == 404

    r = schema.APIResponse.parse_obj(response.json())
    assert r.status == schema.APIStatus.ERROR


def test_api_list_conda_channels_unauth(testclient, seed_conda_store):
    response = testclient.get("api/v1/channel")
    response.raise_for_status()

    r = schema.APIListCondaChannel.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK
    api_channels = set(_.name for _ in r.data)
    assert api_channels == {
        "https://conda.anaconda.org/conda-forge",
    }


def test_api_list_conda_packages_unauth(testclient, seed_conda_store):
    response = testclient.get("api/v1/package")
    response.raise_for_status()

    r = schema.APIListCondaPackage.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK
    assert len(r.data) == 4


# ============ MODIFICATION =============


def test_create_specification_unauth(testclient):
    namespace = "default"
    environment_name = "pytest"

    response = testclient.post(
        "api/v1/specification",
        json={
            "namespace": namespace,
            "specification": json.dumps({"name": environment_name}),
        },
    )
    assert response.status_code == 403

    r = schema.APIResponse.parse_obj(response.json())
    assert r.status == schema.APIStatus.ERROR


# Only testing size values that will always cause errors. Smaller values could
# cause errors as well, but would be flaky since the test conda-store state
# directory might have different lengths on different systems, for instance,
# due to different username lengths.
@pytest.mark.parametrize(
    "size",
    [
        # OSError: [Errno 36] File name too long
        255,
        # OSError: [Errno 36] File name too long
        256,
    ],
)
def test_create_specification_auth_env_name_too_long(
    testclient, celery_worker, authenticate, size
):
    namespace = "default"
    environment_name = "A" * size

    response = testclient.post(
        "api/v1/specification",
        json={
            "namespace": namespace,
            "specification": json.dumps({"name": environment_name}),
        },
    )
    response.raise_for_status()

    r = schema.APIPostSpecification.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK
    build_id = r.data.build_id

    # Try checking that the status is 'FAILED'
    is_updated = False
    for _ in range(5):
        time.sleep(5)

        # check for the given build
        response = testclient.get(f"api/v1/build/{build_id}")
        response.raise_for_status()

        r = schema.APIGetBuild.parse_obj(response.json())
        assert r.status == schema.APIStatus.OK
        assert r.data.specification.name == environment_name
        if r.data.status == "QUEUED":
            continue  # checked too fast, try again
        assert r.data.status == "FAILED"
        assert r.data.status_info == "build_path too long: must be <= 255 characters"
        is_updated = True
        break

    # If we're here, the task didn't update the status on failure
    if not is_updated:
        assert False, "failed to update status"


@pytest.fixture
def win_extended_length_prefix(request):
    # Overrides the attribute before other fixtures are called
    from conda_store_server.app import CondaStore

    assert type(CondaStore.win_extended_length_prefix) is traitlets.Bool
    old_prefix = CondaStore.win_extended_length_prefix
    CondaStore.win_extended_length_prefix = request.param
    yield request.param
    CondaStore.win_extended_length_prefix = old_prefix


@pytest.mark.skipif(sys.platform != "win32", reason="tests a Windows issue")
@pytest.mark.parametrize("win_extended_length_prefix", [True, False], indirect=True)
@pytest.mark.extended_prefix
def test_create_specification_auth_extended_prefix(
    win_extended_length_prefix, testclient, celery_worker, authenticate
):
    # Adds padding to cause an error if the extended prefix is not enabled
    namespace = "default" + "A" * 10
    environment_name = "pytest"

    # The debugpy 1.8.0 package was deliberately chosen because it has long
    # paths internally, which causes issues on Windows due to the path length
    # limit
    response = testclient.post(
        "api/v1/specification",
        json={
            "namespace": namespace,
            "specification": json.dumps(
                {
                    "name": environment_name,
                    "channels": ["conda-forge"],
                    "dependencies": ["debugpy==1.8.0"],
                    "variables": None,
                    "prefix": None,
                    "description": "test",
                }
            ),
        },
        timeout=30,
    )
    response.raise_for_status()

    r = schema.APIPostSpecification.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK
    build_id = r.data.build_id

    # Try checking that the status is 'FAILED'
    is_updated = False
    for _ in range(30):
        time.sleep(5)

        # check for the given build
        response = testclient.get(f"api/v1/build/{build_id}", timeout=30)
        response.raise_for_status()

        r = schema.APIGetBuild.parse_obj(response.json())
        assert r.status == schema.APIStatus.OK
        assert r.data.specification.name == environment_name
        if r.data.status in ("QUEUED", "BUILDING"):
            continue  # checked too fast, try again

        if win_extended_length_prefix:
            assert r.data.status == "COMPLETED"
        else:
            assert r.data.status == "FAILED"
            response = testclient.get(f"api/v1/build/{build_id}/logs", timeout=30)
            response.raise_for_status()
            assert (
                "[WinError 206] The filename or extension is too long" in response.text
            )

        is_updated = True
        break

    # If we're here, the task didn't update the status on failure
    if not is_updated:
        assert False, "failed to update status"


def test_create_specification_auth(testclient, celery_worker, authenticate):
    namespace = "default"
    environment_name = "pytest"

    response = testclient.post(
        "api/v1/specification",
        json={
            "namespace": namespace,
            "specification": json.dumps({"name": environment_name}),
        },
    )
    response.raise_for_status()

    r = schema.APIPostSpecification.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK

    # check for the given build
    response = testclient.get(f"api/v1/build/{r.data.build_id}")
    response.raise_for_status()

    r = schema.APIGetBuild.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK
    assert r.data.specification.name == environment_name

    # check for the given environment
    response = testclient.get(f"api/v1/environment/{namespace}/{environment_name}")
    response.raise_for_status()

    r = schema.APIGetEnvironment.parse_obj(response.json())
    assert r.data.namespace.name == namespace


def test_create_specification_auth_no_namespace_specified(
    testclient, celery_worker, authenticate
):
    namespace = "username"  # same namespace as login
    environment_name = "pytest"

    response = testclient.post(
        "api/v1/specification",
        json={"specification": json.dumps({"name": environment_name})},
    )
    response.raise_for_status()

    r = schema.APIPostSpecification.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK

    # check for the given build
    response = testclient.get(f"api/v1/build/{r.data.build_id}")
    response.raise_for_status()

    r = schema.APIGetBuild.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK
    assert r.data.specification.name == environment_name

    # check for the given environment
    response = testclient.get(f"api/v1/environment/{namespace}/{environment_name}")
    response.raise_for_status()

    r = schema.APIGetEnvironment.parse_obj(response.json())
    assert r.data.namespace.name == namespace


def test_put_build_trigger_build_noauth(testclient, seed_conda_store):
    build_id = 3

    response = testclient.put(f"api/v1/build/{build_id}")
    assert response.status_code == 403

    r = schema.APIResponse.parse_obj(response.json())
    assert r.status == schema.APIStatus.ERROR


def test_put_build_trigger_build_auth(
    testclient, seed_conda_store, authenticate, celery_worker
):
    build_id = 1

    response = testclient.put(f"api/v1/build/{build_id}")
    r = schema.APIPostSpecification.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK

    response = testclient.get(f"api/v1/build/{r.data.build_id}")
    response.raise_for_status()

    r = schema.APIGetBuild.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK


def test_create_namespace_noauth(testclient):
    namespace = "pytest"

    response = testclient.post(f"api/v1/namespace/{namespace}")
    assert response.status_code == 403

    r = schema.APIResponse.parse_obj(response.json())
    assert r.status == schema.APIStatus.ERROR


def test_create_namespace_auth(testclient, authenticate):
    namespace = "pytest"

    response = testclient.post(f"api/v1/namespace/{namespace}")
    response.raise_for_status()

    r = schema.APIAckResponse.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK

    response = testclient.get(f"api/v1/namespace/{namespace}")
    response.raise_for_status()

    r = schema.APIGetNamespace.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK
    assert r.data.name == namespace


def test_create_get_delete_namespace_auth(testclient, celery_worker, authenticate):
    namespace = "pytest-delete-namespace"

    response = testclient.post(f"api/v1/namespace/{namespace}")
    response.raise_for_status()

    r = schema.APIAckResponse.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK

    response = testclient.get(f"api/v1/namespace/{namespace}")
    response.raise_for_status()

    r = schema.APIGetNamespace.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK
    assert r.data.name == namespace

    response = testclient.delete(f"api/v1/namespace/{namespace}")
    response.raise_for_status()

    r = schema.APIAckResponse.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK

    response = testclient.get(f"api/v1/namespace/{namespace}")
    assert response.status_code == 404

    r = schema.APIResponse.parse_obj(response.json())
    assert r.status == schema.APIStatus.ERROR


def test_update_environment_build_unauth(testclient, seed_conda_store):
    namespace = "default"
    name = "name1"
    build_id = 1

    response = testclient.put(
        f"api/v1/environment/{namespace}/{name}", json={"build_id": build_id}
    )
    assert response.status_code == 403

    r = schema.APIResponse.parse_obj(response.json())
    assert r.status == schema.APIStatus.ERROR


def test_update_environment_build_auth(
    testclient, seed_conda_store, authenticate, celery_worker
):
    namespace = "namespace2"
    name = "name4"
    build_id = 4

    response = testclient.put(
        f"api/v1/environment/{namespace}/{name}", json={"build_id": build_id}
    )
    response.raise_for_status()

    r = schema.APIAckResponse.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK

    response = testclient.get(f"api/v1/environment/{namespace}/{name}")
    response.raise_for_status()

    r = schema.APIGetEnvironment.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK
    assert r.data.current_build_id == 4


def test_delete_environment_unauth(testclient, seed_conda_store):
    namespace = "namespace1"
    name = "name3"

    response = testclient.delete(f"api/v1/environment/{namespace}/{name}")
    assert response.status_code == 403

    r = schema.APIResponse.parse_obj(response.json())
    assert r.status == schema.APIStatus.ERROR


def test_delete_environment_auth(
    testclient, seed_conda_store, authenticate, celery_worker
):
    namespace = "namespace1"
    environment_name = "name3"

    response = testclient.post(
        "api/v1/specification",
        json={
            "namespace": namespace,
            "specification": json.dumps({"name": environment_name}),
        },
    )
    response.raise_for_status()

    r = schema.APIAckResponse.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK

    response = testclient.delete(f"api/v1/environment/{namespace}/{environment_name}")
    response.raise_for_status()

    r = schema.APIAckResponse.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK


def test_delete_build_unauth(testclient, seed_conda_store):
    build_id = 4

    response = testclient.delete(f"api/v1/build/{build_id}")
    assert response.status_code == 403

    r = schema.APIResponse.parse_obj(response.json())
    assert r.status == schema.APIStatus.ERROR


def test_delete_build_auth(testclient, seed_conda_store, authenticate, celery_worker):
    build_id = 4

    response = testclient.put(f"api/v1/build/{build_id}")
    r = schema.APIPostSpecification.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK

    new_build_id = r.data.build_id

    response = testclient.get(f"api/v1/build/{new_build_id}")
    response.raise_for_status()

    r = schema.APIGetBuild.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK

    # currently you cannot delete a build before it succeeds or fails
    # realistically you should be able to delete a build
    response = testclient.delete(f"api/v1/build/{new_build_id}")
    assert response.status_code == 400

    # r = schema.APIAckResponse.parse_obj(response.json())
    # assert r.status == schema.APIStatus.OK

    # response = testclient.get(f'api/v1/build/{new_build_id}')
    # assert response.status_code == 404

    # r = schema.APIResponse.parse_obj(response.json())
    # assert r.status == schema.APIStatus.ERROR


def test_prometheus_metrics(testclient):
    response = testclient.get("metrics")
    d = {
        line.split()[0]: line.split()[1]
        for line in response.content.decode("utf-8").split("\n")
    }
    assert {
        "conda_store_disk_free",
        "conda_store_disk_total",
        "conda_store_disk_usage",
    } <= d.keys()


def test_celery_stats(testclient, celery_worker):
    response = testclient.get("celery")
    assert response.json().keys() == {
        "active_tasks",
        "availability",
        "registered_tasks",
        "scheduled_tasks",
        "stats",
    }


@pytest.mark.parametrize(
    "route",
    [
        "/api/v1/setting/",
        "/api/v1/setting/namespace1/",
        "/api/v1/setting/namespace1/name1",
    ],
)
def test_get_settings_unauth(testclient, route):
    response = testclient.get(route)
    assert response.status_code == 403

    r = schema.APIResponse.parse_obj(response.json())
    assert r.status == schema.APIStatus.ERROR


@pytest.mark.parametrize(
    "route",
    [
        "/api/v1/setting/",
        "/api/v1/setting/namespace1/",
        "/api/v1/setting/namespace1/name1",
    ],
)
def test_put_settings_unauth(testclient, route):
    response = testclient.put(route, json={"conda_included_packages": ["numpy"]})
    assert response.status_code == 403

    r = schema.APIResponse.parse_obj(response.json())
    assert r.status == schema.APIStatus.ERROR


@pytest.mark.parametrize(
    "route",
    [
        "/api/v1/setting/",
        "/api/v1/setting/namespace1/",
        "/api/v1/setting/namespace1/name1",
    ],
)
def test_get_settings_auth(testclient, authenticate, route):
    response = testclient.get(route)
    assert response.status_code == 200

    r = schema.APIGetSetting.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK
    assert {
        "default_namespace",
        "filesystem_namespace",
        "default_uid",
        "default_gid",
        "default_permissions",
        "storage_threshold",
        "conda_command",
        "conda_platforms",
        "conda_max_solve_time",
        "conda_indexed_channels",
        "build_artifacts_kept_on_deletion",
        "conda_channel_alias",
        "conda_default_channels",
        "conda_allowed_channels",
        "conda_default_packages",
        "conda_required_packages",
        "conda_included_packages",
        "pypi_default_packages",
        "pypi_required_packages",
        "pypi_included_packages",
        "build_artifacts",
        "default_docker_base_image",
    } <= r.data.keys()


@pytest.mark.parametrize(
    "route",
    [
        "/api/v1/setting/",
    ],
)
def test_put_global_settings_auth(testclient, authenticate, route):
    response = testclient.put(
        route,
        json={
            "conda_max_solve_time": 60 * 60,  # 1 hour
        },
    )
    response.raise_for_status()

    r = schema.APIPutSetting.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK

    response = testclient.get(route)
    response.raise_for_status()
    r = schema.APIPutSetting.parse_obj(response.json())
    assert r.data["conda_max_solve_time"] == 60 * 60


@pytest.mark.parametrize(
    "route",
    [
        "/api/v1/setting/",
        "/api/v1/setting/namespace1",
        "/api/v1/setting/namespace1/name1",
    ],
)
def test_put_environment_settings_auth(testclient, authenticate, route):
    response = testclient.put(
        route, json={"conda_included_packages": ["numpy", "ipykernel"]}
    )
    response.raise_for_status()

    r = schema.APIPutSetting.parse_obj(response.json())
    assert r.status == schema.APIStatus.OK

    response = testclient.get(route)
    response.raise_for_status()
    r = schema.APIPutSetting.parse_obj(response.json())
    assert r.data["conda_included_packages"] == ["numpy", "ipykernel"]


@pytest.mark.parametrize(
    "route",
    [
        "/api/v1/setting/",
        "/api/v1/setting/namespace1",
        "/api/v1/setting/namespace1/name1",
    ],
)
def test_put_environment_settings_auth_invliad_type(testclient, authenticate, route):
    response = testclient.put(route, json={"conda_included_packages": 1})
    assert response.status_code == 400

    r = schema.APIPutSetting.parse_obj(response.json())
    assert r.status == schema.APIStatus.ERROR
    assert "Invalid parsing" in r.message


@pytest.mark.parametrize(
    "route",
    [
        "/api/v1/setting/",
        "/api/v1/setting/namespace1/",
        "/api/v1/setting/namespace1/name1/",
    ],
)
def test_put_settings_auth_invalid_keys(testclient, authenticate, route):
    """This is a test that you cannot set invalid settings"""
    response = testclient.put(
        route,
        json={
            "invalidkey": 1,
            "conda_included_packages": ["numpy"],
        },
    )
    assert response.status_code == 400

    r = schema.APIPutSetting.parse_obj(response.json())
    assert r.status == schema.APIStatus.ERROR
    assert "Invalid setting keys" in r.message


@pytest.mark.parametrize(
    "route",
    [
        "/api/v1/setting/namespace1",
        "/api/v1/setting/namespace1/name1",
    ],
)
def test_put_global_settings_auth_in_namespace_environment(
    testclient, authenticate, route
):
    """This is a test that you cannot set a global setting at the
    namespace or environment settings level"""
    response = testclient.put(
        route,
        json={
            "conda_max_solve_time": 60 * 60,  # 1 hour
        },
    )
    assert response.status_code == 400

    r = schema.APIPutSetting.parse_obj(response.json())
    assert r.status == schema.APIStatus.ERROR
    assert "global setting" in r.message


def test_default_conda_store_dir():
    # Checks the default value of CONDA_STORE_DIR on different platforms
    dir = str(CONDA_STORE_DIR)
    user = os.environ.get("USER") or os.environ.get("USERNAME")
    if sys.platform == "darwin":
        assert dir == f"/Users/{user}/Library/Application Support/conda-store"
    elif sys.platform == "win32":
        assert dir == rf"C:\Users\{user}\AppData\Local\conda-store\conda-store"
    else:
        assert dir == f"/home/{user}/.local/share/conda-store"



"""add worker

Revision ID: 0f7e23ff24ee
Revises: 771180018e1b
Create Date: 2023-12-13 21:01:45.546591

"""

import sqlalchemy as sa

from alembic import op


# revision identifiers, used by Alembic.
revision = "0f7e23ff24ee"
down_revision = "771180018e1b"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "worker",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("initialized", sa.Boolean(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )


def downgrade():
    op.drop_table("worker")



import json
import os
import pathlib
import typing

import yaml

from conda_lock.conda_lock import run_lock

from conda_store_server import action, conda_utils, schema, utils
from conda_store_server.action.utils import logged_command


@action.action
def action_solve_lockfile(
    context,
    conda_command: str,
    specification: schema.CondaSpecification,
    platforms: typing.List[str] = [conda_utils.conda_platform()],
    # Avoids package compatibility issues, see:
    # https://docs.conda.io/projects/conda/en/latest/user-guide/tasks/manage-channels.html
    conda_flags: str = "--strict-channel-priority",
):
    environment_filename = pathlib.Path.cwd() / "environment.yaml"
    lockfile_filename = pathlib.Path.cwd() / "conda-lock.yaml"

    with environment_filename.open("w") as f:
        json.dump(specification.dict(), f)

    # The info command can be used with either mamba or conda
    logged_command(context, [conda_command, "info"])
    # The config command is not supported by mamba
    logged_command(context, ["conda", "config", "--show"])
    logged_command(context, ["conda", "config", "--show-sources"])

    # conda-lock ignores variables defined in the specification, so this code
    # gets the value of CONDA_OVERRIDE_CUDA and passes it to conda-lock via
    # the with_cuda parameter, see:
    # https://github.com/conda-incubator/conda-store/issues/719
    # https://docs.conda.io/projects/conda/en/latest/user-guide/tasks/manage-virtual.html#overriding-detected-packages
    # TODO: Support all variables once upstream fixes are made to conda-lock,
    # see the discussion in issue 719.
    if specification.variables is not None:
        cuda_version = specification.variables.get("CONDA_OVERRIDE_CUDA")
    else:
        cuda_version = None

    # CONDA_FLAGS is used by conda-lock in conda_solver.solve_specs_for_arch
    try:
        conda_flags_name = "CONDA_FLAGS"
        print(f"{conda_flags_name}={conda_flags}")
        os.environ[conda_flags_name] = conda_flags

        run_lock(
            environment_files=[environment_filename],
            platforms=platforms,
            lockfile_path=lockfile_filename,
            conda_exe=conda_command,
            with_cuda=cuda_version,
        )
    finally:
        os.environ.pop(conda_flags_name, None)

    with lockfile_filename.open() as f:
        return yaml.safe_load(f)


@action.action
def action_save_lockfile(
    context,
    specification: schema.LockfileSpecification,
):
    # Note: this calls dict on specification so that the version field is
    # part of the output
    lockfile = specification.dict()["lockfile"]
    lockfile_filename = pathlib.Path.cwd() / "conda-lock.yaml"

    with lockfile_filename.open("w") as f:
        json.dump(lockfile, f, cls=utils.CustomJSONEncoder)

    return lockfile



import os
import pathlib
import shutil
import tempfile
import typing

# This import is needed to avoid the following error on conda imports:
# AttributeError: 'Logger' object has no attribute 'trace'
import conda.gateways.logging  # noqa
import conda_package_handling.api
import conda_package_streaming.url
import filelock

from conda.base.constants import PACKAGE_CACHE_MAGIC_FILE
from conda.common.path import expand, strip_pkg_extension
from conda.core.package_cache_data import (
    PackageCacheRecord,
    PackageRecord,
    getsize,
    read_index_json,
    write_as_json_to_file,
)
from conda.gateways.disk.update import touch

from conda_store_server import action, conda_utils


@action.action
def action_fetch_and_extract_conda_packages(
    context,
    conda_lock_spec: typing.Dict,
    pkgs_dir: pathlib.Path,
    platforms: typing.List[str] = [conda_utils.conda_platform(), "noarch"],
):
    """Download packages from a conda-lock specification using filelocks"""
    packages_searched = 1
    total_packages = len(conda_lock_spec["package"])

    for package in conda_lock_spec["package"]:
        packages_searched += 1
        if package["manager"] == "conda" and package["platform"] in platforms:
            url = package["url"]
            filename = pathlib.Path(url).name
            lock_filename = pkgs_dir / f"{filename}.lock"
            file_path = pkgs_dir / filename
            count_message = f"{packages_searched} of {total_packages}"
            with filelock.FileLock(str(lock_filename)):
                # This magic file, which is currently set to "urls.txt", is used
                # to check cache permissions in conda, see _check_writable in
                # PackageCacheData.
                #
                # Sometimes this file is not yet created while this action is
                # running. Without this magic file, PackageCacheData cache
                # query functions, like query_all, will return nothing.
                #
                # If the magic file is not present, this error might be thrown
                # during the lockfile install action:
                #
                # File "/opt/conda/lib/python3.10/site-packages/conda/misc.py", line 110, in explicit
                #   raise AssertionError("No package cache records found")
                #
                # The code below is from create_package_cache_directory in
                # conda, which creates the package cache, but we only need the
                # magic file part here:
                cache_magic_file = pkgs_dir / PACKAGE_CACHE_MAGIC_FILE
                if not cache_magic_file.exists():
                    sudo_safe = expand(pkgs_dir).startswith(expand("~"))
                    touch(cache_magic_file, mkdir=True, sudo_safe=sudo_safe)

                if file_path.exists():
                    context.log.info(f"SKIPPING {filename} | FILE EXISTS\n")
                else:
                    with tempfile.TemporaryDirectory() as tmp_dir:
                        tmp_dir = pathlib.Path(tmp_dir)
                        file_path = tmp_dir / filename
                        file_path_str = str(file_path)
                        extracted_dir = pathlib.Path(
                            strip_pkg_extension(file_path_str)[0]
                        )
                        extracted_dir_str = str(extracted_dir)
                        context.log.info(f"DOWNLOAD {filename} | {count_message}\n")
                        (
                            filename,
                            conda_package_stream,
                        ) = conda_package_streaming.url.conda_reader_for_url(url)
                        with file_path.open("wb") as f:
                            shutil.copyfileobj(conda_package_stream, f)
                        conda_package_handling.api.extract(
                            file_path_str, dest_dir=extracted_dir
                        )

                        # This code is needed to avoid failures when building in
                        # parallel while using the shared cache.
                        #
                        # Package tarballs contain the info/index.json file,
                        # which is used by conda to create the
                        # info/repodata_record.json file. The latter is used to
                        # interact with the cache. _make_single_record from
                        # PackageCacheData would create the repodata json file
                        # if it's not present, which would happen in conda-store
                        # during the lockfile install action. The repodata file
                        # is not created if it already exists.
                        #
                        # The code that does that in conda is similar to the code
                        # below. However, there is an important difference. The
                        # code in conda would fail to read the url and return None
                        # here:
                        #
                        #   url = self._urls_data.get_url(package_filename)
                        #
                        # And that would result in the channel field of the json
                        # file being set to "<unknown>". This is a problem because
                        # the channel is used when querying cache entries, via
                        # match_individual from MatchSpec, which would always result
                        # in a mismatch because the proper channel value is
                        # different.
                        #
                        # That would make conda think that the package is not
                        # available in the cache, so it would try to download it
                        # outside of this action, where no locking is implemented.
                        #
                        # As of now, conda's cache is not atomic, so the same
                        # dependencies requested by different builds would overwrite
                        # each other causing random failures during the build
                        # process.
                        #
                        # To avoid this problem, the code below does what the code
                        # in conda does but also sets the url properly, which would
                        # make the channel match properly during the query process
                        # later. So no dependencies would be downloaded outside of
                        # this action and cache corruption is prevented.
                        #
                        # To illustrate, here's a diff of an old conda entry, which
                        # didn't work, versus the new one created by this action:
                        #
                        # --- /tmp/old.txt        2024-02-05 01:08:16.879751010 +0100
                        # +++ /tmp/new.txt        2024-02-05 01:08:02.919319887 +0100
                        # @@ -2,7 +2,7 @@
                        #    "arch": "x86_64",
                        #    "build": "conda_forge",
                        #    "build_number": 0,
                        # -  "channel": "<unknown>",
                        # +  "channel": "https://conda.anaconda.org/conda-forge/linux-64",
                        #    "constrains": [],
                        #    "depends": [],
                        #    "features": "",
                        # @@ -15,5 +15,6 @@
                        #    "subdir": "linux-64",
                        #    "timestamp": 1578324546067,
                        #    "track_features": "",
                        # +  "url": "https://conda.anaconda.org/conda-forge/linux-64/_libgcc_mutex-0.1-conda_forge.tar.bz2",
                        #    "version": "0.1"
                        #  }
                        #
                        # Also see the comment above about the cache magic file.
                        # Without the magic file, cache queries would fail even if
                        # repodata_record.json files have proper channels specified.

                        # This file is used to parse cache records via PackageCacheRecord in conda
                        repodata_file = extracted_dir / "info" / "repodata_record.json"

                        raw_json_record = read_index_json(extracted_dir)
                        fn = os.path.basename(file_path_str)
                        md5 = package["hash"]["md5"]
                        size = getsize(file_path_str)

                        package_cache_record = PackageCacheRecord.from_objects(
                            raw_json_record,
                            url=url,
                            fn=fn,
                            md5=md5,
                            size=size,
                            package_tarball_full_path=file_path_str,
                            extracted_package_dir=extracted_dir_str,
                        )

                        repodata_record = PackageRecord.from_objects(
                            package_cache_record
                        )
                        write_as_json_to_file(repodata_file, repodata_record)

                        # This is to ensure _make_single_record in conda never
                        # sees the extracted package directory without our
                        # repodata_record file being there. Otherwise, conda
                        # would attempt to create the repodata file, with the
                        # channel field set to "<unknown>", which would make the
                        # above code pointless. Using symlinks here would be
                        # better since those are atomic on Linux, but I don't
                        # want to create any permanent directories on the
                        # filesystem.
                        shutil.rmtree(pkgs_dir / extracted_dir.name, ignore_errors=True)
                        shutil.move(extracted_dir, pkgs_dir / extracted_dir.name)
                        shutil.move(file_path, pkgs_dir / file_path.name)



import typing

from conda.models.dist import Dist

from conda_store_server import action, api, conda_utils


def list_lockfile_packages(conda_lock_spec: typing.Dict):
    conda_packages = []
    platform = conda_utils.conda_platform()

    for package in conda_lock_spec["package"]:
        if package["manager"] == "conda" and package["platform"] == platform:
            dist = Dist.from_string(package["url"])
            conda_packages.append(
                {
                    "name": dist.name,
                    "build": dist.build,
                    "build_number": dist.build_number,
                    "constrains": None,
                    "depends": [],
                    "license": None,
                    "license_family": None,
                    "size": -1,
                    "subdir": dist.subdir,
                    "timestamp": None,
                    "version": dist.version,
                    "channel_id": dist.base_url,
                    "md5": package["hash"].get("md5"),
                    "sha256": package["hash"].get("sha256", ""),
                    "summary": None,
                    "description": None,
                }
            )
    return conda_packages


@action.action
def action_add_lockfile_packages(
    context, db, conda_lock_spec: typing.Dict, solve_id: int
):
    solve = api.get_solve(db, solve_id=solve_id)
    packages = list_lockfile_packages(conda_lock_spec)

    for package in packages:
        conda_package_build = api.create_or_ignore_conda_package(db, package)
        if conda_package_build is None:
            continue  # pypi package
        solve.package_builds.append(conda_package_build)
        db.commit()



"""add installer

Revision ID: 57cd11b949d5
Revises: 0f7e23ff24ee
Create Date: 2024-01-28 14:31:35.723505

"""

import sqlalchemy as sa

from alembic import op


# revision identifiers, used by Alembic.
revision = "57cd11b949d5"
down_revision = "0f7e23ff24ee"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table(
        "build_artifact",
        schema=None,
    ) as batch_op:
        batch_op.alter_column(
            "artifact_type",
            existing_type=sa.VARCHAR(length=18),
            type_=sa.VARCHAR(length=21),
            existing_nullable=False,
        )
    if not str(op.get_bind().engine.url).startswith("sqlite"):
        op.execute("DROP TYPE IF EXISTS buildartifacttype")


def downgrade():
    op.execute(
        "DELETE FROM build_artifact WHERE artifact_type = 'CONSTRUCTOR_INSTALLER'"
    )
    with op.batch_alter_table(
        "build_artifact",
        schema=None,
    ) as batch_op:
        batch_op.alter_column(
            "artifact_type",
            existing_type=sa.VARCHAR(length=21),
            type_=sa.VARCHAR(length=18),
            existing_nullable=False,
        )



from datetime import datetime, timedelta, timezone


def get_current_time() -> datetime:
    """Get the current time."""
    return datetime.now(timezone.utc)


def get_time_in_future(hours: int) -> datetime:
    """Get the time in the future."""
    current_time = get_current_time()
    future_time = current_time + timedelta(hours=hours)
    return future_time


def get_iso8601_time(hours: int) -> str:
    """Get the time in the future in ISO 8601 format."""
    future_time = get_time_in_future(hours)
    iso_format = future_time.isoformat()
    return iso_format



import pytest

from conda_store_server import api
from conda_store_server.orm import NamespaceRoleMapping
from conda_store_server.utils import BuildPathError


def test_namespace_crud(db):
    namespace_name = "pytest-namespace"

    # starts with no namespaces for test
    assert len(api.list_namespaces(db).all()) == 0

    # create namespace
    namespace = api.create_namespace(db, name=namespace_name)
    db.commit()

    # check that only one namespace exists
    assert len(api.list_namespaces(db).all()) == 1

    # check that ensuring a namespace doesn't create a new one
    api.ensure_namespace(db, name=namespace_name)

    assert len(api.list_namespaces(db).all()) == 1

    # check that getting namespace works
    namespace = api.get_namespace(db, id=namespace.id)
    assert namespace is not None

    # check that deleting a namespace works
    api.delete_namespace(db, id=namespace.id)
    db.commit()

    assert len(api.list_namespaces(db).all()) == 0

    # check that ensuring a namespace doesn't creates one
    api.ensure_namespace(db, name=namespace_name)

    assert len(api.list_namespaces(db).all()) == 1


def test_namespace_role_mapping(db):
    namespace_name = "pytest-namespace"

    # starts with no namespaces for test
    assert len(api.list_namespaces(db).all()) == 0

    # create namespace
    namespace = api.create_namespace(db, name=namespace_name)
    db.commit()

    # check that only one namespace exists
    assert len(api.list_namespaces(db).all()) == 1

    # create a Role Mapping, with a failing entity
    with pytest.raises(Exception):
        NamespaceRoleMapping(
            namespace=namespace, namespace_id=namespace.id, entity="invalid_entity_name"
        )

    # Creates role mappings with valid entity names
    NamespaceRoleMapping(namespace=namespace, namespace_id=namespace.id, entity="org/*")

    NamespaceRoleMapping(
        namespace=namespace, namespace_id=namespace.id, entity="*/team"
    )

    NamespaceRoleMapping(
        namespace=namespace, namespace_id=namespace.id, entity="org/team"
    )

    NamespaceRoleMapping(namespace=namespace, namespace_id=namespace.id, entity="*/*")


@pytest.mark.parametrize(
    "editor_role",
    [
        "editor",
        "developer",
    ],
)
def test_namespace_role_mapping_v2(db, editor_role):
    namespace_name = "pytest-namespace"
    other_namespace_name1 = "pytest-other-namespace1"
    other_namespace_name2 = "pytest-other-namespace2"
    other_namespace_name3 = "pytest-other-namespace3"

    # Starts with no namespaces
    assert len(api.list_namespaces(db).all()) == 0

    # Creates namespaces
    api.create_namespace(db, name=namespace_name)
    api.create_namespace(db, name=other_namespace_name1)
    api.create_namespace(db, name=other_namespace_name2)
    api.create_namespace(db, name=other_namespace_name3)
    db.commit()

    # Checks that all namespaces exist
    assert len(api.list_namespaces(db).all()) == 4

    # Creates role mappings
    api.create_namespace_role(
        db, name=namespace_name, other=other_namespace_name1, role="admin"
    )
    api.create_namespace_role(
        db, name=namespace_name, other=other_namespace_name2, role="admin"
    )
    api.create_namespace_role(
        db, name=namespace_name, other=other_namespace_name3, role="viewer"
    )
    db.commit()

    # Attempts to create a role mapping with an invalid role
    with pytest.raises(ValueError, match=r"invalid role=invalid-role"):
        api.create_namespace_role(
            db, name=namespace_name, other=other_namespace_name3, role="invalid-role"
        )
        db.commit()

    # Attempts to create a role mapping violating the uniqueness constraint
    with pytest.raises(
        Exception,
        match=(
            r"UNIQUE constraint failed: "
            r"namespace_role_mapping_v2.namespace_id, "
            r"namespace_role_mapping_v2.other_namespace_id"
        ),
    ):
        # Runs in a nested transaction since a constraint violation will cause a rollback
        with db.begin_nested():
            api.create_namespace_role(
                db, name=namespace_name, other=other_namespace_name2, role=editor_role
            )
            db.commit()

    # Updates a role mapping
    api.update_namespace_role(
        db, name=namespace_name, other=other_namespace_name2, role=editor_role
    )
    db.commit()

    # Gets all role mappings
    roles = api.get_namespace_roles(db, namespace_name)
    db.commit()
    assert len(roles) == 3

    assert roles[0].id == 1
    assert roles[0].namespace == namespace_name
    assert roles[0].other_namespace == other_namespace_name1
    assert roles[0].role == "admin"

    assert roles[1].id == 2
    assert roles[1].namespace == namespace_name
    assert roles[1].other_namespace == other_namespace_name2
    assert roles[1].role == "developer"  # always developer in the DB

    assert roles[2].id == 3
    assert roles[2].namespace == namespace_name
    assert roles[2].other_namespace == other_namespace_name3
    assert roles[2].role == "viewer"

    # Gets other role mappings
    roles = api.get_other_namespace_roles(db, other_namespace_name1)
    db.commit()
    assert len(roles) == 1
    roles = api.get_other_namespace_roles(db, namespace_name)
    db.commit()
    assert len(roles) == 0

    # Deletes one role mapping
    api.delete_namespace_role(db, name=namespace_name, other=other_namespace_name2)
    db.commit()

    # Gets all role mappings again
    roles = api.get_namespace_roles(db, namespace_name)
    db.commit()
    assert len(roles) == 2

    assert roles[0].id == 1
    assert roles[0].namespace == namespace_name
    assert roles[0].other_namespace == other_namespace_name1
    assert roles[0].role == "admin"

    assert roles[1].id == 3
    assert roles[1].namespace == namespace_name
    assert roles[1].other_namespace == other_namespace_name3
    assert roles[1].role == "viewer"

    # Deletes all role mappings
    api.delete_namespace_roles(db, name=namespace_name)
    db.commit()

    # Checks that roles were deleted
    roles = api.get_namespace_roles(db, name=namespace_name)
    db.commit()
    assert len(roles) == 0


def test_environment_crud(db):
    namespace_name = "pytest-namespace"
    environment_name = "pytest-environment"
    description = "Hello World"

    namespace = api.ensure_namespace(db, name=namespace_name)

    assert len(api.list_environments(db).all()) == 0

    # create environment
    environment = api.create_environment(
        db,
        namespace_id=namespace.id,
        name=environment_name,
        description=description,
    )
    db.commit()

    # check that only one environment exists
    assert len(api.list_environments(db).all()) == 1

    # ensure environment
    api.ensure_environment(db, name=environment_name, namespace_id=namespace.id)

    assert len(api.list_environments(db).all()) == 1

    # check that getting environment works
    environment = api.get_environment(
        db, namespace_id=namespace.id, name=environment_name
    )
    assert environment is not None

    # # check that deleting a environment works
    # api.delete_environment(conda_store.db, namespace_id=namespace.id, name=environment_name)
    # conda_store.db.commit()

    # assert len(api.list_environment(conda_store.db).all()) == 0

    # # check that ensuring a environment doesn't creates one
    # api.ensure_environment(conda_store.db, name=environment_name, namespace_id=namespace.id)

    # assert len(api.list_environments(conda_store.db).all()) == 1


def test_get_set_keyvaluestore(db):
    setting_1 = {"a": 1, "b": 2}
    setting_2 = {"c": 1, "d": 2}
    setting_3 = {"e": 1, "f": 2}

    api.set_kvstore_key_values(db, "pytest", setting_1)
    api.set_kvstore_key_values(db, "pytest/1", setting_2)
    api.set_kvstore_key_values(db, "pytest/1/2", setting_3)

    assert setting_1 == api.get_kvstore_key_values(db, "pytest")
    assert setting_2 == api.get_kvstore_key_values(db, "pytest/1")
    assert setting_3 == api.get_kvstore_key_values(db, "pytest/1/2")

    # test updating a prefix
    api.set_kvstore_key_values(db, "pytest", setting_2)
    assert {**setting_1, **setting_2} == api.get_kvstore_key_values(db, "pytest")

    # test updating a prefix
    api.set_kvstore_key_values(db, "pytest", {"c": 999, "d": 999}, update=False)
    assert {**setting_1, **setting_2} == api.get_kvstore_key_values(db, "pytest")


def test_build_path_too_long(db, conda_store, simple_specification):
    conda_store.store_directory = "A" * 800
    build_id = conda_store.register_environment(
        db, specification=simple_specification, namespace="pytest"
    )
    build = api.get_build(db, build_id=build_id)
    with pytest.raises(
        BuildPathError, match=r"build_path too long: must be <= 255 characters"
    ):
        build.build_path(conda_store)



import logging
import os
import posixpath
import sys
import time

from enum import Enum
from threading import Thread

import uvicorn

from fastapi import FastAPI, HTTPException, Request
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import FileResponse, JSONResponse, RedirectResponse
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from sqlalchemy.pool import QueuePool
from starlette.middleware.sessions import SessionMiddleware
from traitlets import (
    Bool,
    Dict,
    Instance,
    Integer,
    List,
    Type,
    Unicode,
    default,
    validate,
)
from traitlets.config import Application, catch_config_error

import conda_store_server
import conda_store_server.dbutil as dbutil

from conda_store_server import __version__, orm, storage
from conda_store_server.app import CondaStore
from conda_store_server.server import auth, views


class _Color(str, Enum):
    GREEN = "\x1b[32m"
    RED = "\x1b[31m"
    RESET = "\x1b[0m"


class CondaStoreServer(Application):
    aliases = {
        "config": "CondaStoreServer.config_file",
    }

    flags = {
        "standalone": (
            {
                "CondaStoreServer": {
                    "standalone": True,
                }
            },
            "Run conda-store-server in standalone mode with celery worker as a subprocess of webserver",
        ),
    }

    reload = Bool(
        False,
        help="Enable reloading on code change",
        config=True,
    )

    log_level = Integer(
        logging.INFO,
        help="log level to use",
        config=True,
    )

    enable_ui = Bool(True, help="serve the web ui for conda-store", config=True)

    enable_api = Bool(
        True,
        help="enable the rest api for conda-store",
        config=True,
    )

    enable_registry = Bool(
        True, help="enable the docker registry for conda-store", config=True
    )

    enable_metrics = Bool(
        True,
        help="enable the prometheus metrics for conda-store",
        config=True,
    )

    address = Unicode(
        "0.0.0.0", help="ip address or hostname for conda-store server", config=True
    )

    port = Integer(8080, help="port for conda-store server", config=True)

    registry_external_url = Unicode(
        "localhost:8080",
        help='external hostname and port to access docker registry cannot contain "http://" or "https://"',
        config=True,
    )

    url_prefix = Unicode(
        "/",
        help="the prefix URL (subdirectory) for the entire application; "
        "it MUST start with a forward slash - tip: "
        "use this to run conda-store within an existing website.",
        config=True,
    )

    config_file = Unicode(
        help="config file to load for conda-store",
        config=True,
    )

    behind_proxy = Bool(
        False,
        help="Indicates if server is behind web reverse proxy such as Nginx, Traefik, Apache. Will use X-Forward-.. headers to detemine scheme. Do not set to true if not behind proxy since Flask will trust any X-Forward-... header",
        config=True,
    )

    templates = Instance(
        help="Initialized fastapi.templating.Jinja2Templates to use for html templates",
        klass=Jinja2Templates,
        config=True,
    )

    template_vars = Dict(
        {},
        help="Extra variables to be passed into jinja templates for page rendering",
        config=True,
    )

    additional_routes = List(
        [],
        help="Additional routes for conda-store to serve [(path, method, function), ...]",
        config=True,
    )

    cors_allow_origins = List(
        [], help="list of allowed origins for CORS requests", config=True
    )

    cors_allow_origin_regex = Unicode(
        ".*",
        help="regex string to match against origins that should be permitted to make cross-origin requests",
        config=True,
        allow_none=True,
    )

    @default("templates")
    def _default_templates(self):
        import conda_store_server.server

        templates_directory = os.path.join(
            os.path.dirname(conda_store_server.server.__file__), "templates"
        )
        return Jinja2Templates(directory=templates_directory)

    @validate("config_file")
    def _validate_config_file(self, proposal):
        if not os.path.isfile(proposal.value):
            print(
                "ERROR: Failed to find specified config file: {}".format(
                    proposal.value
                ),
                file=sys.stderr,
            )
            sys.exit(1)
        return proposal.value

    authentication_class = Type(
        default_value=auth.DummyAuthentication,
        klass=auth.Authentication,
        allow_none=False,
        config=True,
    )

    max_page_size = Integer(
        100, help="maximum number of items to return in a single page", config=True
    )

    standalone = Bool(
        False,
        help="Run application in standalone mode with workers running as subprocess",
        config=True,
    )

    @catch_config_error
    def initialize(self, *args, **kwargs):
        super().initialize(*args, **kwargs)
        self.load_config_file(self.config_file)

        self.conda_store = CondaStore(parent=self, log=self.log)

        self.conda_store.ensure_directories()
        self.log.info(
            f"Running conda-store with database: {self.conda_store.database_url}"
        )
        self.log.info(
            f"Running conda-store with store directory: {self.conda_store.store_directory}"
        )

        if self.conda_store.upgrade_db:
            dbutil.upgrade(self.conda_store.database_url)

        self.authentication = self.authentication_class(
            parent=self,
            log=self.log,
            authentication_db=self.conda_store.session_factory,
        )

        # ensure checks on redis_url
        self.conda_store.redis_url

    def init_fastapi_app(self):
        def trim_slash(url):
            return url[:-1] if url.endswith("/") else url

        app = FastAPI(
            title="conda-store",
            version=__version__,
            openapi_url=posixpath.join(self.url_prefix, "openapi.json"),
            docs_url=posixpath.join(self.url_prefix, "docs"),
            redoc_url=posixpath.join(self.url_prefix, "redoc"),
            contact={
                "name": "Quansight",
                "url": "https://quansight.com",
            },
            license_info={
                "name": "BSD 3-Clause",
                "url": "https://opensource.org/licenses/BSD-3-Clause",
            },
        )

        app.add_middleware(
            CORSMiddleware,
            allow_origins=self.cors_allow_origins,
            allow_origin_regex=self.cors_allow_origin_regex,
            allow_credentials=True,
            allow_headers=["*"],
            allow_methods=["*"],
        )
        app.add_middleware(
            SessionMiddleware, secret_key=self.authentication.authentication.secret
        )

        # ensure that template variables are inserted into templates
        self.templates.env.globals.update(self.template_vars)

        @app.middleware("http")
        async def conda_store_middleware(request: Request, call_next):
            request.state.conda_store = self.conda_store
            request.state.server = self
            request.state.authentication = self.authentication
            request.state.templates = self.templates
            response = await call_next(request)
            return response

        @app.exception_handler(HTTPException)
        async def http_exception_handler(request, exc):
            return JSONResponse(
                {
                    "status": "error",
                    "message": exc.detail,
                },
                status_code=exc.status_code,
            )

        # Prints exceptions to the terminal
        # https://fastapi.tiangolo.com/tutorial/handling-errors/#re-use-fastapis-exception-handlers
        # https://github.com/tiangolo/fastapi/issues/1241
        @app.exception_handler(Exception)
        async def exception_handler(request, exc):
            print(exc)
            return await http_exception_handler(request, exc)

        app.include_router(
            self.authentication.router,
            prefix=trim_slash(self.url_prefix),
        )

        if self.enable_api:
            app.include_router(
                views.router_api,
                prefix=trim_slash(self.url_prefix),
            )

        if self.enable_registry:
            # docker registry api specification does not support a url_prefix
            app.include_router(views.router_registry)

        if self.enable_ui:
            app.include_router(
                views.router_ui,
                prefix=trim_slash(self.url_prefix) + "/admin",
            )

            app.include_router(
                views.router_conda_store_ui,
                prefix=trim_slash(self.url_prefix),
            )

            # serving static files
            import conda_store_server.server

            app.mount(
                trim_slash(self.url_prefix) + "/static/",
                StaticFiles(
                    directory=os.path.join(
                        os.path.dirname(conda_store_server.server.__file__),
                        "static",
                    ),
                ),
                name="static",
            )

            # convenience to redirect "/" to home page when using a prefix
            # realistically this url will not be hit with a proxy + prefix
            if self.url_prefix != "/":

                @app.get("/")
                def redirect_home(request: Request):
                    return RedirectResponse(request.url_for("get_conda_store_ui"))

            @app.get("/favicon.ico", include_in_schema=False)
            async def favicon():
                return FileResponse(
                    os.path.join(
                        os.path.dirname(conda_store_server.server.__file__),
                        "static",
                        "favicon.ico",
                    )
                )

        if self.enable_metrics:
            app.include_router(
                views.router_metrics,
                prefix=trim_slash(self.url_prefix),
            )

        if self.additional_routes:
            for path, method, func in self.additional_routes:
                getattr(app, method)(path, name=func.__name__)(func)

        if isinstance(self.conda_store.storage, storage.LocalStorage):
            self.conda_store.storage.storage_url = (
                f"{trim_slash(self.url_prefix)}/storage"
            )
            app.mount(
                self.conda_store.storage.storage_url,
                StaticFiles(directory=self.conda_store.storage.storage_path),
                name="static-storage",
            )

        return app

    def _check_worker(self, delay=5):
        # Creates a new DB connection since this will be run in a separate
        # thread and connections cannot be shared between threads
        session_factory = orm.new_session_factory(
            url=self.conda_store.database_url,
            poolclass=QueuePool,
        )

        # Waits in a loop for the worker to become ready, which is
        # communicated by the worker via task_initialize_worker
        while True:
            with session_factory() as db:
                q = db.query(orm.Worker).first()
                if q is not None and q.initialized:
                    self.log.info(
                        f"{_Color.GREEN}" "Worker initialized" f"{_Color.RESET}"
                    )
                    break

            time.sleep(delay)
            self.log.warning(
                f"{_Color.RED}"
                "Waiting for worker... "
                "Use --standalone if running outside of docker"
                f"{_Color.RESET}"
            )

    def start(self):
        fastapi_app = self.init_fastapi_app()

        with self.conda_store.session_factory() as db:
            self.conda_store.ensure_settings(db)
            self.conda_store.ensure_namespace(db)
            self.conda_store.ensure_conda_channels(db)

            # This ensures the database has no Worker table entries when the
            # server starts, which is necessary for the worker to signal that
            # it's ready via task_initialize_worker. Old Worker entries could
            # still be in the database on startup after they were added on the
            # previous run and the server was terminated.
            #
            # Note that this cleanup is deliberately done on startup because the
            # server could be terminated due to a power failure, which would
            # leave no chance for cleanup actions to run on shutdown.
            #
            # The database is used for worker-server communication because it
            # will work regardless of celery_broker_url used, which can be Redis
            # or just point to a database connection.
            db.query(orm.Worker).delete()
            db.commit()

        # We cannot check whether the worker is ready right away and block. When
        # running via docker, the worker is started *after* the server is
        # running because it relies on config files created by the server.
        # So we just keep checking in a separate thread until the worker is
        # ready.
        worker_checker = Thread(target=self._check_worker)
        worker_checker.start()

        # start worker if in standalone mode
        if self.standalone:
            import multiprocessing

            multiprocessing.set_start_method("spawn")

            from conda_store_server.worker.app import CondaStoreWorker

            process = multiprocessing.Process(target=CondaStoreWorker.launch_instance)
            process.start()

        try:
            # Note: the logger needs to be defined here for the output to show
            # up, self.log doesn't work here either
            logger = logging.getLogger("app")
            logger.addHandler(logging.StreamHandler())
            logger.setLevel(self.log_level)
            logger.info(f"Starting server on {self.address}:{self.port}")

            uvicorn.run(
                fastapi_app,
                host=self.address,
                port=self.port,
                workers=1,
                proxy_headers=self.behind_proxy,
                forwarded_allow_ips=("*" if self.behind_proxy else None),
                reload=self.reload,
                reload_dirs=(
                    [os.path.dirname(conda_store_server.__file__)]
                    if self.reload
                    else []
                ),
            )
        except:
            import traceback

            traceback.print_exc()
            raise
        finally:
            if self.standalone:
                process.join()
            worker_checker.join()



import pathlib

from conda_store_server import action, conda_utils, utils


@action.action
def action_get_conda_prefix_stats(context, conda_prefix: pathlib.Path):
    # safeguard to try and prevent destructive actions
    if not conda_utils.is_conda_prefix(conda_prefix):
        raise ValueError("given prefix is not a conda environment")

    stats = {}
    stats["disk_usage"] = int(utils.disk_usage(conda_prefix))
    return stats



"""add build hash

Revision ID: e17b4cc6e086
Revises: 03c839888c82
Create Date: 2024-03-26 04:39:24.275214

"""

import sqlalchemy as sa

from alembic import op


# revision identifiers, used by Alembic.
revision = "e17b4cc6e086"
down_revision = "03c839888c82"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column("build", sa.Column("hash", sa.Unicode(length=32), nullable=True))


def downgrade():
    op.drop_column("build", "hash")



import pathlib

from conda_store_server import action


@action.action
def action_generate_conda_docker(
    context,
    conda_prefix: pathlib.Path,
    default_docker_image: str,
    container_registry,
    output_image_name: str,
    output_image_tag: str,
):
    raise RuntimeError(
        "Generating Docker images is currently not supported, see "
        "https://github.com/conda-incubator/conda-store/issues/666"
    )

    # Import is inside the function because conda_docker is only available on
    # Linux
    from conda_docker.conda import (
        build_docker_environment_image,
        conda_info,
        fetch_precs,
        find_user_conda,
        precs_from_environment_prefix,
    )

    user_conda = find_user_conda()
    info = conda_info(user_conda)
    download_dir = info["pkgs_dirs"][0]
    precs = precs_from_environment_prefix(str(conda_prefix), download_dir, user_conda)
    records = fetch_precs(download_dir, precs)
    base_image = container_registry.pull_image(default_docker_image)
    image = build_docker_environment_image(
        base_image=base_image,
        output_image=f"{output_image_name}:{output_image_tag}",
        records=records,
        default_prefix=info["env_vars"]["CONDA_ROOT"],
        download_dir=download_dir,
        user_conda=user_conda,
        channels_remap=info.get("channels_remap", []),
        layering_strategy="single",
    )
    return image



import hashlib
import json
import os
import pathlib

from conda.core.prefix_data import PrefixData

from conda_store_server import action, api


def list_conda_prefix_packages(conda_prefix: pathlib.Path):
    """
    Returns a list of the packages that exist for a given prefix

    """
    packages = []

    prefix_data = PrefixData(str(conda_prefix))
    prefix_data.load()

    for record in prefix_data.iter_records():
        package = {
            "build": record.build,
            "build_number": record.build_number,
            "constrains": list(record.constrains),
            "depends": list(record.depends),
            "license": record.license,
            "license_family": record.license_family,
            "md5": hashlib.md5(
                open(record.package_tarball_full_path, "rb").read()
            ).hexdigest(),
            "sha256": hashlib.sha256(
                open(record.package_tarball_full_path, "rb").read()
            ).hexdigest(),
            "name": record.name,
            "size": getattr(record, "size", 0),
            "subdir": record.subdir,
            "timestamp": record.timestamp,
            "version": record.version,
            "channel_id": record.channel.base_url,
            "summary": None,
            "description": None,
        }

        info_json = os.path.join(record.extracted_package_dir, "info/about.json")
        if os.path.exists(info_json):
            info = json.load(open(info_json))
            package["summary"] = info.get("summary")
            package["description"] = info.get("description")

        packages.append(package)
    return packages


@action.action
def action_add_conda_prefix_packages(
    context,
    db,
    conda_prefix: pathlib.Path,
    build_id: int,
):
    build = api.get_build(db, build_id=build_id)
    packages = list_conda_prefix_packages(conda_prefix)

    for package in packages:
        conda_package_build = api.create_or_ignore_conda_package(db, package)
        if conda_package_build is None:
            continue  # pypi package
        build.package_builds.append(conda_package_build)
        db.commit()



import datetime
import uuid

import pytest

from conda_store_server.schema import AuthenticationToken, Permissions
from conda_store_server.server.auth import (
    AuthenticationBackend,
    RBACAuthorizationBackend,
)


@pytest.mark.parametrize(
    "expression, resource, match",
    [
        ("e*/d*", "example/data", True),
        ("e*/d*", "example/eta", False),
        ("*e*d*/*e", "deed/rate", True),
    ],
)
def test_compile_arn_regex(expression, resource, match):
    regex = RBACAuthorizationBackend.compile_arn_regex(expression)
    assert (regex.match(resource) is not None) == match


@pytest.mark.parametrize(
    "expression, namespace, name",
    [
        ("e*/d*", "e%", "d%"),
        ("*e*d*/*e", "%e%d%", "%e"),
    ],
)
def test_compile_arn_sql_like(expression, namespace, name):
    result = RBACAuthorizationBackend.compile_arn_sql_like(expression)
    assert (namespace, name) == result


@pytest.mark.parametrize(
    "token_string,authenticated",
    [
        ("", False),
        ("asdf", False),
    ],
)
def test_authenticated(token_string, authenticated):
    authentication = AuthenticationBackend()
    authentication.secret = "supersecret"
    assert (authentication.authenticate(token_string) is not None) == authenticated


def test_valid_token():
    authentication = AuthenticationBackend()
    authentication.secret = "supersecret"

    token = authentication.encrypt_token(
        AuthenticationToken(
            primary_namespace="default",
            role_bindings={
                "default/*": ["viewer"],
                "e*/e*": ["admin"],
            },
        )
    )

    token_model = authentication.authenticate(token)
    assert isinstance(token_model, AuthenticationToken)


def test_expired_token():
    authentication = AuthenticationBackend()
    authentication.secret = "supersecret"

    token = authentication.encrypt_token(
        AuthenticationToken(
            primary_namespace="default",
            exp=datetime.datetime.utcnow() - datetime.timedelta(hours=1),
            role_bindings={
                "default/*": ["viewer"],
                "e*/e*": ["admin"],
            },
        )
    )

    assert authentication.authenticate(token) is None


@pytest.mark.parametrize(
    "entity_bindings,arn,permissions,authorized",
    [
        (
            {
                "example-namespace/example-name": {"developer"},
            },
            "example-namespace/example-name",
            {Permissions.ENVIRONMENT_CREATE},
            True,
        ),
        (
            {
                "example-namespace/example-name": {"developer", "viewer"},
            },
            "example-namespace/example-name",
            {Permissions.ENVIRONMENT_DELETE},
            False,
        ),
        (
            {
                "example-namespace/example-name": {"developer", "admin"},
            },
            "example-namespace/example-name",
            {Permissions.ENVIRONMENT_DELETE},
            True,
        ),
        (
            {
                "e*/e*am*": {"admin"},
            },
            "example-namespace/example-name",
            {Permissions.ENVIRONMENT_CREATE},
            True,
        ),
        (
            {
                "e*/e*am*": {"viewer"},
            },
            "example-namespace/example-name",
            {Permissions.ENVIRONMENT_CREATE},
            False,
        ),
    ],
)
def test_authorization(conda_store, entity_bindings, arn, permissions, authorized):
    authorization = RBACAuthorizationBackend(
        authentication_db=conda_store.session_factory
    )

    entity = AuthenticationToken(
        primary_namespace="example_namespace", role_bindings=entity_bindings
    )

    assert authorized == authorization.authorize(entity, arn, permissions)


_viewer_permissions = {
    Permissions.ENVIRONMENT_READ,
    Permissions.NAMESPACE_READ,
    Permissions.NAMESPACE_ROLE_MAPPING_READ,
}
_editor_permissions = {
    Permissions.BUILD_CANCEL,
    Permissions.ENVIRONMENT_CREATE,
    Permissions.ENVIRONMENT_READ,
    Permissions.ENVIRONMENT_UPDATE,
    Permissions.ENVIRONMENT_SOLVE,
    Permissions.NAMESPACE_READ,
    Permissions.NAMESPACE_ROLE_MAPPING_READ,
    Permissions.SETTING_READ,
}
_admin_permissions = {
    Permissions.BUILD_DELETE,
    Permissions.BUILD_CANCEL,
    Permissions.ENVIRONMENT_CREATE,
    Permissions.ENVIRONMENT_DELETE,
    Permissions.ENVIRONMENT_READ,
    Permissions.ENVIRONMENT_UPDATE,
    Permissions.ENVIRONMENT_SOLVE,
    Permissions.NAMESPACE_CREATE,
    Permissions.NAMESPACE_DELETE,
    Permissions.NAMESPACE_READ,
    Permissions.NAMESPACE_UPDATE,
    Permissions.NAMESPACE_ROLE_MAPPING_CREATE,
    Permissions.NAMESPACE_ROLE_MAPPING_READ,
    Permissions.NAMESPACE_ROLE_MAPPING_UPDATE,
    Permissions.NAMESPACE_ROLE_MAPPING_DELETE,
    Permissions.SETTING_READ,
    Permissions.SETTING_UPDATE,
}


@pytest.mark.parametrize(
    "role, permissions",
    [
        ("viewer", _viewer_permissions),
        ("developer", _editor_permissions),
        ("editor", _editor_permissions),
        ("admin", _admin_permissions),
    ],
)
@pytest.mark.parametrize("role_mappings_version", [1, 2])
def test_end_to_end_auth_flow_v1(
    conda_store_server,
    testclient,
    authenticate,
    role_mappings_version,
    role,
    permissions,
):
    # Configures authentication
    namespace = f"this-{uuid.uuid4()}"
    other_namespace = f"other-{uuid.uuid4()}"

    conda_store = conda_store_server.conda_store

    authentication = AuthenticationBackend()
    authentication.secret = "supersecret"

    token = authentication.encrypt_token(
        AuthenticationToken(
            primary_namespace=namespace,
            # No default roles
            role_bindings={},
        )
    )

    token_model = authentication.authenticate(token)

    authorization = RBACAuthorizationBackend(
        authentication_db=conda_store.session_factory,
        role_mappings_version=role_mappings_version,
    )

    def authorize():
        return authorization.authorize(
            AuthenticationToken(
                primary_namespace=token_model.primary_namespace,
                role_bindings=token_model.role_bindings,
            ),
            f"{other_namespace}/example-name",
            permissions,
        )

    # No default roles
    assert authorize() is False

    # Creates new namespaces
    for n in (namespace, other_namespace):
        response = testclient.post(f"api/v1/namespace/{n}")
        response.raise_for_status()

    # Deletes roles to start with a clean state
    response = testclient.put(
        f"api/v1/namespace/{namespace}", json={"role_mappings": {}}
    )
    response.raise_for_status()

    # Creates role for 'namespace' with access to 'other_namespace'
    response = testclient.put(
        f"api/v1/namespace/{namespace}",
        json={
            "role_mappings": {
                f"{other_namespace}/ex*-name": [role],
            }
        },
    )
    response.raise_for_status()

    # Should succeed now if v1
    if role_mappings_version == 1:
        assert authorize() is True
    else:
        assert authorize() is False

    # Deletes created roles
    response = testclient.put(
        f"api/v1/namespace/{namespace}", json={"role_mappings": {}}
    )
    response.raise_for_status()

    # Should fail again
    assert authorize() is False


@pytest.mark.parametrize(
    "role, permissions",
    [
        ("viewer", _viewer_permissions),
        ("developer", _editor_permissions),
        ("editor", _editor_permissions),
        ("admin", _admin_permissions),
    ],
)
@pytest.mark.parametrize("role_mappings_version", [1, 2])
def test_end_to_end_auth_flow_v2(
    conda_store_server,
    testclient,
    authenticate,
    role_mappings_version,
    role,
    permissions,
):
    # Configures authentication
    namespace = f"this-{uuid.uuid4()}"
    other_namespace = f"other-{uuid.uuid4()}"

    conda_store = conda_store_server.conda_store

    authentication = AuthenticationBackend()
    authentication.secret = "supersecret"

    token = authentication.encrypt_token(
        AuthenticationToken(
            primary_namespace=namespace,
            # No default roles
            role_bindings={},
        )
    )

    token_model = authentication.authenticate(token)

    authorization = RBACAuthorizationBackend(
        authentication_db=conda_store.session_factory,
        role_mappings_version=role_mappings_version,
    )

    def authorize():
        return authorization.authorize(
            AuthenticationToken(
                primary_namespace=token_model.primary_namespace,
                role_bindings=token_model.role_bindings,
            ),
            f"{other_namespace}/example-name",
            permissions,
        )

    # No default roles
    assert authorize() is False

    # Creates new namespaces
    for n in (namespace, other_namespace):
        response = testclient.post(f"api/v1/namespace/{n}")
        response.raise_for_status()

    # Deletes roles to start with a clean state
    response = testclient.delete(f"api/v1/namespace/{other_namespace}/roles")
    response.raise_for_status()

    # Creates role for 'namespace' with access to 'other_namespace'
    response = testclient.post(
        f"api/v1/namespace/{other_namespace}/role",
        json={"other_namespace": namespace, "role": role},
    )
    response.raise_for_status()

    # Should succeed now if v2
    if role_mappings_version == 2:
        assert authorize() is True
    else:
        assert authorize() is False

    # Deletes created roles
    response = testclient.delete(f"api/v1/namespace/{other_namespace}/roles")
    response.raise_for_status()

    # Should fail again
    assert authorize() is False


@pytest.mark.parametrize(
    "arn_1,arn_2,value",
    [
        ("ab/cd", "a*b/c*d", True),
        ("a1111b/c22222d", "a*b/c*d", True),
        ("a1/cd", "a*b/c*d", False),
        ("abc/ed", "a*b*c/e*d", True),
        ("a111bc/ed", "a*b*c/e*d", True),
        ("a111b2222c/e3333d", "a*b*c/e*d", True),
        ("aaabbbcccc/eeddd", "a*b*c/e*d", True),
        ("aaabbbcccc1/eeddd", "a*b*c/e*d", False),
        ("aaabbbcccc1c/eeddd", "a*b*c/e*d", True),
    ],
)
def test_is_arn_subset(arn_1, arn_2, value):
    assert RBACAuthorizationBackend.is_arn_subset(arn_1, arn_2) == value


@pytest.mark.parametrize(
    # "entity_bindings, new_entity_bindings, authenticated, value",
    "entity_bindings, new_entity_bindings, value",
    [
        # */* viewer is a subset of admin
        (
            {"*/*": ["admin"]},
            {"*/*": ["viewer"]},
            # False,
            True,
        ),
        (
            {"*/*": ["admin"]},
            {"*/*": ["viewer"]},
            # True,
            True,
        ),
        # */* admin is not a subset of viewer
        (
            {"*/*": ["viewer"]},
            {"*/*": ["admin"]},
            # False,
            False,
        ),
        (
            {"*/*": ["viewer"]},
            {"*/*": ["admin"]},
            # True,
            False,
        ),
        # a/b viewer is a subset of admin
        (
            {"a/b": ["admin"]},
            {"a/b": ["viewer"]},
            # False,
            True,
        ),
        (
            {"a/b": ["admin"]},
            {"a/b": ["viewer"]},
            # True,
            True,
        ),
        # a/b admin is not a subset of viewer
        (
            {"a/b": ["viewer"]},
            {"a/b": ["admin"]},
            # False,
            False,
        ),
        (
            {"a/b": ["viewer"]},
            {"a/b": ["admin"]},
            # True,
            False,
        ),
        # default/* vs. */*
        (
            {"*/*": ["viewer"]},
            {"default/*": ["viewer"]},
            # False,
            True,
        ),
        (
            {"*/*": ["viewer"]},
            {"default/*": ["viewer"]},
            # True,
            True,
        ),
        # efault/* vs. d*/*
        (
            {"d*/*": ["viewer"]},
            {"efault/*": ["viewer"]},
            # False,
            False,
        ),
        (
            {"d*/*": ["viewer"]},
            {"efault/*": ["viewer"]},
            # True,
            False,
        ),
        # multiple entities keys
        (
            {"d*/*": ["viewer"], "de*/*": ["admin"]},
            {"default/*": ["developer"]},
            # False,
            True,
        ),
        (
            {"d*/*": ["viewer"], "de*/*": ["admin"]},
            {"default/*": ["developer"]},
            # True,
            True,
        ),
        # multiple entities keys
        (
            {"d*/*": ["viewer"], "de*/*": ["admin"]},
            {"dcefault/*": ["developer"]},
            # False,
            False,
        ),
        (
            {"d*/*": ["viewer"], "de*/*": ["admin"]},
            {"dcefault/*": ["developer"]},
            # True,
            False,
        ),
        # multiple entities keys
        (
            {"d*/*": ["viewer"]},
            {"d*/*": ["viewer"], "dc*/*": ["viewer"]},
            # False,
            True,
        ),
        (
            {"d*/*": ["viewer"]},
            {"d*/*": ["viewer"], "dc*/*": ["viewer"]},
            # True,
            True,
        ),
        # multiple entities keys
        (
            {"d*/*": ["viewer"]},
            {"d*/*": ["viewer"], "dc*/*": ["developer"]},
            # False,
            False,
        ),
        (
            {"d*/*": ["viewer"]},
            {"d*/*": ["viewer"], "dc*/*": ["developer"]},
            # True,
            False,
        ),
    ],
)
def test_is_subset_entity_permissions(
    conda_store,
    entity_bindings,
    new_entity_bindings,
    # authenticated,
    value,
):
    authorization = RBACAuthorizationBackend(
        authentication_db=conda_store.session_factory
    )

    entity = AuthenticationToken(role_bindings=entity_bindings)
    new_entity = AuthenticationToken(role_bindings=new_entity_bindings)

    assert authorization.is_subset_entity_permissions(entity, new_entity) == value



"""add canceled status

Revision ID: 03c839888c82
Revises: 57cd11b949d5
Create Date: 2024-01-29 03:56:36.889909

"""

import sqlalchemy as sa

from alembic import op


# revision identifiers, used by Alembic.
revision = "03c839888c82"
down_revision = "57cd11b949d5"
branch_labels = None
depends_on = None


# Migrating from/to VARCHAR having the same length might look strange, but it
# serves a purpose. This will be a no-op in SQLite because it represents Python
# enums as VARCHAR, but it will convert the enum in PostgreSQL to VARCHAR. The
# old type is set to VARCHAR here because you can cast an enum to VARCHAR, which
# is needed for the migration to work. In the end, both DBs will use VARCHAR to
# represent the Python enum, which makes it easier to support both DBs at the
# same time.
def upgrade():
    with op.batch_alter_table(
        "build",
        schema=None,
    ) as batch_op:
        batch_op.alter_column(
            "status",
            existing_type=sa.VARCHAR(length=9),
            type_=sa.VARCHAR(length=9),
            existing_nullable=False,
        )
    if not str(op.get_bind().engine.url).startswith("sqlite"):
        op.execute("DROP TYPE IF EXISTS buildstatus")


def downgrade():
    # There are foreign key constraints linking build ids to other tables. So
    # just mark the builds as failed, which was the status previously used for
    # canceled builds
    op.execute("UPDATE build SET status = 'FAILED' WHERE status = 'CANCELED'")



import json
import pathlib

from conda_store_server import action, schema


@action.action
def action_install_specification(
    context,
    conda_command: str,
    specification: schema.CondaSpecification,
    conda_prefix: pathlib.Path,
):
    environment_filename = pathlib.Path.cwd() / "environment.yaml"
    with environment_filename.open("w") as f:
        json.dump(specification.dict(), f)

    command = [
        conda_command,
        "env",
        "create",
        "-p",
        str(conda_prefix),
        "-f",
        str(environment_filename),
    ]

    context.run(command, check=True)



import pathlib
import tempfile

import pytest

from fastapi.templating import Jinja2Templates
from fastapi.testclient import TestClient

from conda_store_server import schema


def test_conda_store_server_enable_ui(conda_store_server):
    conda_store_server.enable_ui = False
    client = TestClient(conda_store_server.init_fastapi_app())
    response = client.get("/")
    assert response.status_code == 404

    conda_store_server.enable_ui = True
    client = TestClient(conda_store_server.init_fastapi_app())
    response = client.get("/")
    assert response.status_code == 200


def test_conda_store_server_enable_api(conda_store_server):
    conda_store_server.enable_api = False
    client = TestClient(conda_store_server.init_fastapi_app())
    response = client.get("/api/v1/")
    assert response.status_code == 404

    conda_store_server.enable_api = True
    client = TestClient(conda_store_server.init_fastapi_app())
    response = client.get("/api/v1/")
    assert response.status_code == 200


def test_conda_store_server_enable_registry(conda_store_server):
    conda_store_server.enable_registry = False
    client = TestClient(conda_store_server.init_fastapi_app())
    response = client.get("/v2/")
    assert response.status_code == 404

    conda_store_server.enable_registry = True
    client = TestClient(conda_store_server.init_fastapi_app())
    response = client.get("/v2/")
    assert response.status_code == 401  # unauthorized


def test_conda_store_server_enable_metrics(conda_store_server):
    conda_store_server.enable_metrics = False
    client = TestClient(conda_store_server.init_fastapi_app())
    response = client.get("/metrics")
    assert response.status_code == 404

    conda_store_server.enable_metrics = True
    client = TestClient(conda_store_server.init_fastapi_app())
    response = client.get("/metrics")
    assert response.status_code == 200


def test_conda_store_server_url_prefix(conda_store_server):
    conda_store_server.url_prefix = "/a/test/prefix/"
    client = TestClient(conda_store_server.init_fastapi_app())
    response = client.get("/", follow_redirects=False)
    assert response.status_code == 307

    response = client.get("/a/test/prefix/", follow_redirects=False)
    assert response.status_code == 200

    conda_store_server.url_prefix = "/"
    client = TestClient(conda_store_server.init_fastapi_app())
    response = client.get("/", follow_redirects=False)
    assert response.status_code == 200


def test_conda_store_server_templates(conda_store_server):
    with tempfile.TemporaryDirectory() as tmpdir:
        tmpdir = pathlib.Path(tmpdir)

        templates = Jinja2Templates(directory=str(tmpdir))

        content = b"<h1>Hello World {{ test_variable }}</h1>"
        with (tmpdir / "home.html").open("wb") as f:
            f.write(content)

        conda_store_server.templates = templates
        conda_store_server.template_vars = {"test_variable": "Test"}
        client = TestClient(conda_store_server.init_fastapi_app())
        response = client.get("/admin/")
        assert response.status_code == 200
        assert response.content == b"<h1>Hello World Test</h1>"


def test_conda_store_server_authentication_class(conda_store_server):
    from conda_store_server.server.auth import DummyAuthentication

    class MyAuthentication(DummyAuthentication):
        password = "helloworld"

    conda_store_server.authentication_class = MyAuthentication
    conda_store_server.initialize()

    client = TestClient(conda_store_server.init_fastapi_app())
    response = client.post(
        "/login/", json={"username": "username", "password": "password"}
    )
    assert response.status_code == 403

    response = client.post(
        "/login/", json={"username": "username", "password": "helloworld"}
    )
    assert response.status_code == 200


def test_conda_store_server_additional_routes(conda_store_server):
    def a_route(a: str, b: str):
        return {"data": f"Hello World {a} {b}"}

    conda_store_server.additional_routes = [
        (
            "/a/path/to/another/",
            "get",
            a_route,
        )
    ]

    client = TestClient(conda_store_server.init_fastapi_app())
    response = client.get("/a/path/to/another/", params={"a": "c", "b": "d"})
    assert response.status_code == 200
    assert response.json() == {"data": "Hello World c d"}


def test_conda_store_settings_conda_channels_packages_validate_valid(db, conda_store):
    conda_store.set_settings(
        db,
        data={
            "conda_allowed_channels": ["conda-forge"],
            "conda_included_packages": ["ipykernel"],
            "conda_required_packages": ["flask"],
            "pypi_included_packages": ["scipy"],
            "pypi_required_packages": ["numpy"],
        },
    )

    global_specification = conda_store.validate_specification(
        db,
        conda_store,
        namespace="default",
        specification=schema.CondaSpecification(
            name="test",
            channels=["conda-forge"],
            dependencies=[
                "flask",
                {"pip": ["numpy"]},
            ],
        ),
    )
    assert global_specification.channels == ["conda-forge"]
    assert sorted(global_specification.dependencies, key=lambda i: str(i)) == [
        "flask",
        "ipykernel",
        schema.CondaSpecificationPip(pip=["numpy", "scipy"]),
    ]

    conda_store.set_settings(
        db,
        namespace="default",
        data={
            "conda_allowed_channels": ["conda-forge"],
            "conda_included_packages": ["ipykernel", "pandas"],
            "conda_required_packages": ["flask"],
            "pypi_included_packages": ["scipy"],
            "pypi_required_packages": ["numpy"],
        },
    )

    namespace_specification = conda_store.validate_specification(
        db,
        conda_store,
        namespace="default",
        specification=schema.CondaSpecification(
            name="test",
            channels=["conda-forge"],
            dependencies=[
                "flask",
                {"pip": ["numpy"]},
            ],
        ),
    )
    assert namespace_specification.channels == ["conda-forge"]
    assert sorted(namespace_specification.dependencies, key=lambda i: str(i)) == [
        "flask",
        "ipykernel",
        "pandas",
        schema.CondaSpecificationPip(pip=["numpy", "scipy"]),
    ]

    conda_store.set_settings(
        db,
        namespace="default",
        environment_name="test",
        data={
            "conda_allowed_channels": ["conda-forge"],
            "conda_included_packages": ["ipykernel", "asdf"],
            "conda_required_packages": ["flask"],
            "pypi_included_packages": ["scipy"],
            "pypi_required_packages": ["numpy"],
        },
    )

    environment_specification = conda_store.validate_specification(
        db,
        conda_store,
        namespace="default",
        specification=schema.CondaSpecification(
            name="test",
            channels=["conda-forge"],
            dependencies=[
                "flask",
                {"pip": ["numpy"]},
            ],
        ),
    )
    assert environment_specification.channels == ["conda-forge"]
    assert sorted(environment_specification.dependencies, key=lambda i: str(i)) == [
        "asdf",
        "flask",
        "ipykernel",
        schema.CondaSpecificationPip(pip=["numpy", "scipy"]),
    ]

    # not allowed channel name
    with pytest.raises(ValueError):
        conda_store.validate_specification(
            db,
            conda_store,
            namespace="default",
            specification=schema.CondaSpecification(
                name="test",
                channels=["bad-channel"],
                dependencies=["flask", {"pip": ["numpy"]}],
            ),
        )

    # missing required conda package
    with pytest.raises(ValueError):
        conda_store.validate_specification(
            db,
            conda_store,
            namespace="default",
            specification=schema.CondaSpecification(
                name="test", channels=["conda-forge"], dependencies=[{"pip": ["numpy"]}]
            ),
        )

    # missing required pip package
    with pytest.raises(ValueError):
        conda_store.validate_specification(
            db,
            conda_store,
            namespace="default",
            specification=schema.CondaSpecification(
                name="test", channels=["conda-forge"], dependencies=["flask"]
            ),
        )
